/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/encryption/Encryption.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc. 
 * All rights reserved.
 */
package com.netapp.common.encryption;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.spec.EncodedKeySpec;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;

/**
 * Wraps basic encryption methods for general use. Callers can encrypt or decrypt messages ased on their
 * supplied key and algorithm. For Asymmetric usage, they can load the public and private keys from files.
 * 
 * @author joshb
 *
 */
public class Encryption {
    private static final NALogger LOGGER = NALogger.getLogger(Encryption.class);
    
    private static final String ALGORITHM_X509 = "X.509";
    
    protected Encryption() {
    }

    /**
     * Using the key and algorithm, encrypt the incoming byte array and return the result.
     * 
     * @param toEncrypt
     * @param encryptionKey
     * @param algorithm
     * @return Encrypted text of toEncrypt
     * @throws EncryptionException
     */
    public static byte [] encrypt(byte [] toEncrypt, Key encryptionKey, String algorithm) throws EncryptionException {
        try {
            Cipher cipher = getCipher(encryptionKey, algorithm, Cipher.ENCRYPT_MODE);

            return cipher.doFinal(toEncrypt);
        }
        catch (IllegalBlockSizeException e) {
            throw new EncryptionException(e);
        }
        catch (BadPaddingException e) {
            throw new EncryptionException(e);
        }
    }
    
    /**
     * Using the key and algorithm, decrypt the incoming byte array and return the result.
     * 
     * @param toDecrypt
     * @param deccryptionKey
     * @param algorithm
     * @return Plaintext of toDecrypt
     * @throws EncryptionException
     */
    public static byte [] decrypt(byte [] toDecrypt, Key deccryptionKey, String algorithm) throws EncryptionException {
        try {
            Cipher cipher = getCipher(deccryptionKey, algorithm, Cipher.DECRYPT_MODE);
            
            return cipher.doFinal(toDecrypt);
        }
        catch (IllegalBlockSizeException e) {
            throw new EncryptionException(e);
        }
        catch (BadPaddingException e) {
            throw new EncryptionException(e);
        }
    }
    
    /**
     * Load a PKCS8 private key from a file.
     * 
     * @param privateKeyFile
     * @param algorithm
     * @return PrivateKey from file
     * @throws EncryptionException
     */
    public static PrivateKey getPrivateKeyFromFile(String privateKeyFile, String algorithm) throws EncryptionException {
        try {
            byte [] privateKeyBytes = readFile(privateKeyFile);
            KeyFactory keyFactory = KeyFactory.getInstance(getAlgorithmFromTransformation(algorithm));
            EncodedKeySpec privateKeySpec = new PKCS8EncodedKeySpec(privateKeyBytes);
            
            return keyFactory.generatePrivate(privateKeySpec);
        }
        catch (NoSuchAlgorithmException e) {
            throw new EncryptionException(e);
        }
        catch (IOException e) {
            throw new EncryptionException(e);
        }
        catch (InvalidKeySpecException e) {
            throw new EncryptionException(e);
        }
    }
    
    /**
     * Load an X.509 Cert public key from a file.
     * 
     * @param publicKeyFile
     * @return X.509 Cert public key
     * @throws EncryptionException 
     */
    public static PublicKey getPublicKeyFromFile(String publicKeyFile) throws EncryptionException {
        try {
            InputStream stream = getFileInputStream(publicKeyFile);
            CertificateFactory factory = CertificateFactory.getInstance(ALGORITHM_X509);
            X509Certificate cert = (X509Certificate) factory.generateCertificate(stream);
            stream.close();
            
            return cert.getPublicKey();
        }
        catch (FileNotFoundException e) {
            throw new EncryptionException(e);
        }
        catch (CertificateException e) {
            throw new EncryptionException(e);
        }
        catch (IOException e) {
            throw new EncryptionException(e);
        }
    }
    
    /**
     * Given a key, algorithm and the cipher mode to use (encrypt/decrypt), construct a new Cipher
     * that has been properly initialized for use.
     *  
     * @param cipherKey
     * @param algorithm
     * @param cipherMode
     * @return Cipher implementing the requested encryption parameters
     * @throws EncryptionException
     */
    public static Cipher getCipher(Key cipherKey, String algorithm, int cipherMode) throws EncryptionException {
        try {
            Cipher cipher = Cipher.getInstance(algorithm);
            cipher.init(cipherMode, cipherKey);

            return cipher;
        }
        catch (NoSuchAlgorithmException e) {
            throw new EncryptionException(e);
        }
        catch (NoSuchPaddingException e) {
            throw new EncryptionException(e);
        }
        catch (InvalidKeyException e) {
            throw new EncryptionException(e);
        }
    }
    
    /**
     * Read the bytes from a file. If the file cannot be found, attempt to locate it on the class path.
     * If it cannot be located on the class path, throw an exception.
     * 
     * @param keyFile
     * @return Contents of keyfile
     * @throws IOException
     */
    protected static byte[] readFile(String keyFile) throws IOException {
        File file = new File(keyFile);
        long length = -1;
        InputStream stream = null;
        
        if (!file.exists() && !file.isFile()) {
            URL url = ClassLoader.getSystemResource(keyFile);
            
            if (null != url) {
                URLConnection connection = url.openConnection();
            
                length = connection.getContentLength();
                stream = connection.getInputStream();
            }
            else {
                // If we did not find it on the classpath, return the original
                throw new FileNotFoundException(LOGGER.error(MsgKey.ENCRYPTION_ERROR_FILE_NOT_FOUND, keyFile));
            }
        }
        else {
            length = file.length();
            stream = new FileInputStream(file.getPath());
        }
        
        return readFileFromInputStream(stream, length);
    }
    
    /**
     * Read from the input stream until we have read all of the bytes up to the given
     * length.
     * 
     * @param stream
     * @param length
     * @return Contents up to length of stream
     */
    protected static byte[] readFileFromInputStream(InputStream stream, long length) {
        ByteArrayOutputStream contents = new ByteArrayOutputStream();
        byte[] buffer = new byte[(int)length];
        int read = 0;
        
        try {
            while (read < length) {
                int inputLength = stream.read(buffer, read, buffer.length - read);
                contents.write(buffer, read, inputLength);
                read += inputLength;
            }
        }
        catch (IOException e) {
            e.printStackTrace();
        }

        return contents.toByteArray();
    }
    
    /**
     * Find the input stream for the file. If the file does not exist on the filesystem, locate
     * it from the classpath.
     * 
     * @see #readFile(String)
     * 
     * @param filepath
     * @return Inputstream for filepath
     * @throws IOException
     */
    protected static InputStream getFileInputStream(String filepath) throws IOException {
        File file = new File(filepath);
        
        if (!file.exists() && !file.isFile()) {
            URL url = ClassLoader.getSystemResource(filepath);
            if (null == url) {
                url = Thread.currentThread().getContextClassLoader().getResource(filepath);
            }
            
            if (null != url) {
                return url.openStream();
            }
            else {
                // If we did not find it on the classpath, return the original
                throw new FileNotFoundException(LOGGER.error(MsgKey.ENCRYPTION_ERROR_FILE_NOT_FOUND, filepath));
            }
        }
        else {
            return new FileInputStream(file);
        }
    }
    
    /**
     * A transformation contains the following form.
     * 
     *   <algorithm>/<mode>/<padding>
     *   
     * For some methods, only the algorithm is needed. This method will strip off the mode and padding.
     * 
     * @param transformation
     * @return Algorithim portion of tranformation
     */
    public static String getAlgorithmFromTransformation(String transformation) {
        /*
         * If the algorithm contains a '/', just use the first part for the KeyFactory.
         */
        int index = transformation.indexOf("/");
        
        if (-1 != index) {
            transformation = transformation.substring(0, index);
        }
        
        return transformation;
    }

}
