/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/OperationContext.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.apache.log4j.MDC;

import com.netapp.common.MsgKey;
import com.netapp.common.flow.api.OperationState;
import com.netapp.common.logging.LogMessage;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.StringUtils;

/**
 * This class represents the entire state of a operation
 * workflow at any point in time. It may be persisted and
 * reloaded at a later point in time.
 * <p>
 * This class is public so that specific applications may subclass
 * this and a corresponding {@link PersistenceManager} to store 
 * additonal fields outside the register interface. It should not
 * be accessed by Tasks in general.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class OperationContext implements Comparable<OperationContext>, Serializable, StackFrame {
    private static final transient NALogger logger = NALogger.getLogger(OperationContext.class);
   
    private static final long serialVersionUID = 1L;
    
    private static ThreadLocal<Stack<OperationContext>> threadContext = new ThreadLocal<Stack<OperationContext>>() {
        @Override
        protected Stack<OperationContext> initialValue() {
            return new Stack<OperationContext>();
        }
    };
    
    private transient PersistenceManager manager;
    private transient OperationTemplate operationTemplate;
    
    private OperationId id;
    private String clientRequestId;
    private Date createTime;
    private OperationState state;
    private Date holdUntilTime;
    private String checkpoint;
    private Date checkpointTime;
    private String abortCheckpoint;
    private int retryCount;
    private boolean cancel;
    private Throwable taskException;
    private Throwable abortException;
    private FailureStrategy failureStrategy;
    
    /**
     * Has this context changed since we loaded it?
     */
    private transient boolean dirty;
    
    /**
     * The set of registers seen by the running task
     */
    private Map<String, Object> registers;
    
    /**
     * The set of registers seen by the persistence framework.
     * This is the set of registers that are serialized out during a checkpoint,
     * which may not be the same as those modified during the current task.
     */
    private Map<String, Object> persistentRegisters;
    
    /**
     * Construct a new operation context with an existing ID.
     * This should be used only to rehydrate a formerly stored Operation.
     * The rest of the fields will need to be filled in as well.
     * @param id
     */
    protected OperationContext(PersistenceManager manager, OperationId id, OperationTemplate operationTemplate) {
        this.manager = manager;
        this.id = id;
        this.operationTemplate = operationTemplate;
        this.dirty = true;
    }
    
    /**
     * Construct a new context without the initial parameters.
     * Subclasses that override {@link #enterContext()} and {@link #exitContext} to
     * add additional information to the NDC or MDC should use this version
     * and call setInitialParameters themselves after configuring any other
     * application specific values.
     * @param manager
     * @param operationTemplate
     */
    protected OperationContext(PersistenceManager manager, OperationTemplate operationTemplate) {
        this.id = new OperationId();
        this.manager = manager;
        this.operationTemplate = operationTemplate;
        this.dirty = true;
        this.createTime = new Date();
        this.state = OperationState.PAUSED;
        this.retryCount = 0;
        this.cancel = false;
        this.failureStrategy = FailureStrategy.IGNORE;
    }
    
    /**
     * Construct a new context. 
     * @param manager
     * @param operationTemplate
     * @param params
     */
    protected OperationContext(PersistenceManager manager, OperationTemplate operationTemplate, Map<String, ?> params) {
        this(manager, operationTemplate);
        setInitialParameters(params);
    }
    
    protected void setInitialParameters(Map<String, ?> params) {
        this.clientRequestId = (String)params.get("clientRequestId");
        this.persistentRegisters = new HashMap<String, Object>();
        
        this.enterContext();
        try {
            if (!StringUtils.isEmpty(this.clientRequestId)) {
                if (OperationId.isId(this.clientRequestId)) {
                    throw new FlowException(logger.error(MsgKey.FLOW_OPERATION_ILLEGAL_CLIENT_REQUEST_ID, this.clientRequestId));
                }
                OperationContext ctx2 = this.manager.getContext(this.clientRequestId);
                if (ctx2 != null) {
                    throw new FlowException(logger.error(MsgKey.FLOW_OPERATION_DUPLICATE_CLIENT_REQUEST_ID, this.clientRequestId, ctx2));
                }
            }
            logger.debug(MsgKey.FLOW_OPERATION_NEW, this.getOperationType());

            for (RegisterTemplate input : this.operationTemplate.getInputs()) {
                if (!params.containsKey(input.getInputName())) {
                    throw new FlowException(logger.error(MsgKey.FLOW_OPERATION_INPUT_NOT_DECLARED, input.getInputName(), this.getOperationType()));
                }
                logger.internal("Setting input " + input.getName());
                this.persistentRegisters.put(input.getName(), params.get(input.getInputName()));
            }
            loadWorkingRegisters();
            this.dirty = true;
        }
        finally {
            this.exitContext();
        }
    }
    
    /**
     * Mark this object as clean.
     * This should only be used by the persistence manager
     * when finished a checkpoint or rehydrating a stored
     * copy.
     *
     */
    protected void markClean() {
        this.dirty = false;
    }
    
    /**
     * Return the id of this context
     * @return context id
     */
    public OperationId getId() {
        return this.id;
    }
    
    public void setClientRequestId(String clientRequestId) {
        assert this.clientRequestId == null;
        this.clientRequestId = clientRequestId;
    }
    
    public String getClientRequestId() {
        return this.clientRequestId;
    }
    
    /**
     * Signal the current thread is about to begin working on this context.
     * This is primarily designed for logging - by calling this method, any
     * context data the application wishes to register with the logger diagnostic
     * contexts can be accomplished.
     *
     */
    public void enterContext() {
        logger.internal("Entering context %s in thread %s", this.toString(), Thread.currentThread().getName());
        OperationContext.threadContext.get().push(this);
        MDC.put(FlowConstants.OPERATION_MDC, this.id.toString());
        MDC.put(FlowConstants.OPERATION_TYPE_MDC, this.operationTemplate.getName());
        if (!StringUtils.isEmpty(this.clientRequestId)) {
            MDC.put(FlowConstants.CLIENT_ID_MDC, this.clientRequestId);
        }
    }
    
    public void exitContext() {
        Stack<OperationContext> stack = OperationContext.threadContext.get();
        if (!stack.empty()) {
            OperationContext ctx = stack.pop();
            MDC.remove(FlowConstants.OPERATION_MDC);
            MDC.remove(FlowConstants.CLIENT_ID_MDC);
            MDC.remove(FlowConstants.OPERATION_TYPE_MDC);
            logger.internal("Left context %s in thread %s", ctx, Thread.currentThread().getName());
            if (this != ctx) {
                logger.internal("Left context %s, expected context %s", ctx, this);
            }
            if (!stack.empty()) {
                stack.peek().enterContext();
            }
        }
    }
    
    public OperationTemplate getOperationTemplate() {
        return this.operationTemplate;
    }

    public String getOperationType() {
        return this.operationTemplate.getName();
    }
    
    public Date getCreateTime() {
        return this.createTime;
    }
    
    /**
     * Set the create time for this context.
     * This may only be done once when rehydrating a saved operation.
     * @param createTime
     */
    protected void setCreateTime(Date createTime) {
        assert this.createTime == null;
        this.dirty = true;
        this.createTime = createTime;
    }
    
    public OperationState getState() {
        return this.state;
    }
    
    public void setState(OperationState state) {
        logger.internal("Operation %s changing state %s to %s", this.id, this.state, state);
        this.dirty = true;
        this.state = state;
    }


    public Date getHoldUntilTime() {
        return holdUntilTime;
    }

    public void setHoldUntilTime(Date holdUntilTime) {
        this.dirty = true;
        this.holdUntilTime = holdUntilTime;
    }

    public String getCheckpoint() {
        return checkpoint;
    }

    public void setCheckpoint(String checkpoint) {
        assert (checkpoint == null || this.operationTemplate.checkpointExists(checkpoint));
        this.dirty = true;
        this.checkpoint = checkpoint;
    }
    
    public Date getCheckpointTime() {
        return this.checkpointTime;
    }
    
    public String getAbortCheckpoint() {
        return abortCheckpoint;
    }

    public void setAbortCheckpoint(String abortCheckpoint) {
        assert (abortCheckpoint == null || this.operationTemplate.abortCheckpointExists(abortCheckpoint));
        this.dirty = true;
        this.abortCheckpoint = abortCheckpoint;
    }

    public boolean isAborting() {
        // CP / ACP Stack Operation
        // N    N   Tasks Tasks[0]
        // Y    X   Tasks Tasks[CP + 1]
        //
        // N    Y   Abort Abort[CP + 1]
        return ((this.state == OperationState.RUNNING || this.state == OperationState.PAUSED)
            && taskException != null);
    }
    
    /**
     * Reset the retry count at a successful checkpoint. 
     * 
     */
    public void resetRetryCount() {
        this.retryCount = 0;
    }
    
    /**
     * Increment the retry count by one. 
     *
     */
    public void incrementRetryCount() {
        this.retryCount++;
    }
    
    /**
     * Retrieve the number of times the current tasklist has been restarted since a checkpoint.
     * @return retry count
     */
    public int getRetryCount() {
        return this.retryCount;
    }
    
    /**
     * Set the number of times the current tasklist has been restarted since a checkpoint.
     * @param retryCount
     */
    public void setRetryCount(int retryCount) {
        this.retryCount = retryCount;
    }
    
    /**
     * @see com.netapp.common.flow.StackFrame#getRegisterValue(java.lang.String)
     */
    public Object getRegisterValue(String name) {
        return registers.get(name);
    }
    
    /**
     * @see com.netapp.common.flow.StackFrame#setRegisterValue(java.lang.String, java.lang.Object)
     */
    public void setRegisterValue(String name, Object obj) {
        this.dirty = true;
        this.registers.put(name, obj);
    }
    
    /**
     * 
     * @see com.netapp.common.flow.StackFrame#getContext()
     */
    public OperationContext getContext() {
        return this;
    }
    
    /**
     * Retrieve the complete set of registers in this context.
     * @return Map of register contents
     */
    public Map<String, Object> getPersistentRegisters() {
        return this.persistentRegisters;
    }
    
    /**
     * Set the complete set of registers. This sets the persistent and working set.
     * @param registers
     */
    protected void setRegisters(Map<String, Object> registers) {
        this.persistentRegisters = new HashMap<String, Object>(registers);
        loadWorkingRegisters();
        this.dirty = true;
    }
    
    /**
     * Copy the persistent registers over to the working set.
     *
     */
    protected void loadWorkingRegisters() {
        assert this.persistentRegisters != null;
        this.registers = new HashMap<String, Object>(this.persistentRegisters);
    }
    
    /**
     * Copy the working registers over to the persistent set prior to initiating
     * a full checkpoint.
     *
     */
    protected void loadPersistentRegisters() {
        assert this.registers != null;
        this.persistentRegisters = new HashMap<String, Object>(this.registers);
    }
    
    public void cancel() {
        this.cancel = true;
    }
    
    protected void setCancel(boolean cancel) {
        this.cancel = cancel;
    }
    
    public boolean isCancelled() {
        return this.cancel;
    }
    
    public Throwable getTaskException() {
        return this.taskException;
    }
    
    public void setTaskException(Throwable t) {
        this.taskException = t;
    }
    
    public Object getClientAccessibleError() {
        if (this.taskException instanceof ClientAccessibleErrorProvider) {
            
            ClientAccessibleErrorProvider errorProvider = (ClientAccessibleErrorProvider) this.taskException; 
            return errorProvider.getClientAccessibleError();
        }
        
        return null;
    }
    
    public Throwable getAbortException() {
        return this.abortException;
    }
    
    public void setAbortException(Throwable t) {
        this.abortException = t;
    }
    
    public boolean isDirty() {
        return this.dirty;
    }
    
    /**
     * Is this operation in a terminal state and also
     * committed to our persistence store.
     * @return true if this operation is in a committed terminal state, false otherwise
     */
    public boolean isTerminalState() {
        return isTerminalState(true);
    }
    
    /**
     * Is this operation in a terminal state.
     * @param persisted When false, the operation must only be in a termial state. If true,
     * the operation must be terminal and committed to the persistence store.
     * @return true if this operation is in a terminal state, false otherwise
     */
    protected boolean isTerminalState(boolean persisted) {
        return (!persisted || !this.dirty)
                && (this.state == OperationState.COMPLETE
                        || this.state == OperationState.FAILED
                        || this.state == OperationState.INCOMPLETE);
    }
    /**
     * Get the final results for this operation to return to an external
     * caller.
     * This may only be called after the operation is in a terminal state
     * @return The final state and any outputs of the executed operation
     */
    public Result getResult() {
        Result result = new Result(this.id);
        result.setType(this.getOperationType());
        result.setCreateTime(this.createTime);
        result.setLastActiveTime(this.checkpointTime);
        result.setClientRequestId(this.clientRequestId);
        result.setClientAccessibleError(getClientAccessibleError());

        // Always return the copy of current messages generated since the
        // workflow started. Create a new list here so we don't get a concurrent
        // modification exception iterating over the list
        List<LogMessage> logs = this.manager.getMessages(this.id);
        if (logs != null) {
            result.messages = new ArrayList<LogMessage>(logs);
        }
        else {
            result.messages = new ArrayList<LogMessage>();
        }
     
        if (!this.isTerminalState()) {
            result.setState(OperationState.RUNNING);
        }
        else {
            result.setState(this.state);

            if (this.state == OperationState.COMPLETE) {
                for (RegisterTemplate output : this.operationTemplate.getOutputs()) {
                    result.addOutput(output.getOutputName(), registers.get(output.getName()));
                }
            }
        }
        
        return result;
    }
    
    public void persist() {
        this.checkpointTime = new Date();
        this.manager.saveContext(this);
    }

    @Override
    public int hashCode() {
        return this.id.hashCode();
    }
    
    public int compareTo(OperationContext o) {
        return this.id.compareTo(o.id);
    }
    
    @Override
    public boolean equals(Object o) {
        if (o instanceof OperationContext) {
            return this.id.equals(((OperationContext)o).id);
        }
        return false;
    }
    
    @Override
    public String toString() {
        return this.id.toString();
    }

    /**
     * @return Returns the failureStrategy.
     */
    public FailureStrategy getFailureStrategy() {
        return failureStrategy;
    }

    /**
     * @param failureStrategy The failureStrategy to set.
     */
    public void setFailureStrategy(FailureStrategy failureStrategy) {
        this.failureStrategy = failureStrategy;
    }
    
    /**
     * Acquire a lock for the current context.
     * 
     * @param keys Key values to lock
     * @throws UnableToAcquireLockException 
     * @see TaskUtils
     */
    public void acquireLocks(Set<String> keys) throws UnableToAcquireLockException {
        this.manager.acquireLocks(this.id, keys);
    }
    
    /**
     * Retrieve the context associated to the current thread, set in {@link #enterContext()}
     * @return The operation context of the current thread, or null if no current context is set
     */
    public static OperationContext getThreadContext() {
        OperationContext ctx = null;
        if (!OperationContext.threadContext.get().empty()) {
            ctx = OperationContext.threadContext.get().peek();
        }
        return ctx;
    }

    /**
     * Compare the supplied persistence manager to our
     * manager.
     * @param manager2
     * @return True if the supplied manager 
     */
    boolean isManagedBy(PersistenceManager manager2) {
        return this.manager == manager2;
    }

}
