/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/Threadpool.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.AtomicLong;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;

/**
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
class Threadpool extends ThreadGroup implements Comparable<Threadpool>, ThreadpoolMBean {
    private static final NALogger logger = NALogger.getLogger(Threadpool.class);
    
    private final int size;
    
    private Engine engine;
    /** Idle time in seconds between scheduler wakeups */
    private int idleTime = FlowConstants.DEFAULT_THREADPOOL_IDLE; 
    private ScheduleThread scheduler;
    
    private OperationThread[] pool;
    private BlockingQueue<OperationThread> idleThreads = new LinkedBlockingQueue<OperationThread>();
    
    private Date lastOperationStartTime;
    private Date lastOperationStopTime;
    private AtomicLong operationStartCount = new AtomicLong();
    
    /**
     * The scheduler uses this to determine what we can run.
     * We use it frequently, so we don't want to generate it each time.
     */
    private final Set<String> validOperations = new HashSet<String>();
    
    Threadpool(String name, int size) {
        super(name);
        this.size = size;
        pool = new OperationThread[this.size];
        for (int i = 0; i < this.size; i++) {
            OperationThread thread = new OperationThread(i);
            pool[i] = thread;
        }
        
        logger.internal("Defined thread pool %s (%d)", this.getName(), this.size);
    }
    
    public void setIdleTime(int idleTime) {
        this.idleTime = idleTime;
    }
    public Set<String> getValidOperations() {
        return this.validOperations;
    }
    
    void addValidOperation(String operationName) {
        logger.internal("Assigning task %s to pool %s", operationName, this.getName());
        validOperations.add(operationName);
    }
    
    public int getSize() {
        return this.size;
    }
    
    public int getActiveCount() {
        int active = 0;
        for (OperationThread thread : pool) {
            active += thread.isIdle() ? 0 : 1;
        }
        return active;
    }
    
    public int getIdleCount() {
        return idleThreads.size();
    }
    
    public String[] getThreads() {
        String[] titles = new String[this.size];
        int i = 0;
        for (OperationThread thread : pool) {
            titles[i++] = thread.getName();
        }
        return titles;
    }
    
    public Date getLastOperationStartTime() {
        return this.lastOperationStartTime;
    }
    
    public Date getLastOperationStopTime() {
        return this.lastOperationStopTime;
    }
    
    public long getOperationStartCount() {
        return this.operationStartCount.longValue();
    }
    
    public void notifyScheduler() {
        synchronized(this.scheduler) {
            this.scheduler.notify();
        }
    }
    
    void start(Engine engine) {
        assert this.scheduler == null;
        assert this.engine == null;
        this.engine = engine;
        this.scheduler = new ScheduleThread();
        this.scheduler.start();
        
        for (OperationThread thread : pool) {
            // Threads put themselves in the idle queue
            // as soon as they startup:
            thread.start();
        }
    }
    
    void shutdown() {        
        while (this.scheduler.isAlive()) {
            this.scheduler.shutdown();
            try {
                this.scheduler.join(1000);
            }
            catch (InterruptedException e) { }
            
        }

        for (OperationThread thread : this.pool) {
            while (thread.isAlive()) {
                thread.shutdown();
                try {
                    thread.join(1000);
                }
                catch (InterruptedException e) { }
            }
        }
        
        /*
         * See burt 335334
         * Need to call this.destroy() to remove the reference to ourself from our parent ThreadGroup
         */
        logger.internal("Destroying threadpool %s", this.getName());
        this.destroy();
    }
    
    public int compareTo(Threadpool o) {
        return this.getName().compareTo(o.getName());
    }

    private class ScheduleThread extends Thread {
        private volatile boolean shutdown;
        
        public ScheduleThread() {
            super(Threadpool.this, Threadpool.this.getName() + " SCHEDULER");
            this.setDaemon(true);
        }
                
        @Override
        public void run() {
            TaskUtils.setThreadEngine(Threadpool.this.engine);
            this.shutdown = false;
            
            while (!this.shutdown) {
                try {
                    // Retrieve a worker thread off the queue. This will block until one becomes available
                    OperationThread worker = idleThreads.take();
                    
                    OperationContext context = null;
                    synchronized(worker) {
                        assert worker.operation == null;
                        logger.internal("Threadpool %s polling for new contexts", Threadpool.this.getName());
                        try {
                            context = Threadpool.this.engine.getPersistenceManager().getRunnableContext(Threadpool.this.validOperations);
                            if (context != null) {
                                context.enterContext();
                                logger.debug(MsgKey.FLOW_OPERATION_STARTING, Threadpool.this.getName());
                                Operation operation = new Operation(Threadpool.this.engine, context);
                                worker.setOperation(operation);
                                context.exitContext();
                            }
                        }
                        
                        catch (FlowException e) {
                            logger.error(MsgKey.FLOW_OPERATION_SCHEDULER_EXCEPTION, Threadpool.this.getName(), e.getLocalizedMessage());
                        }
                        // Wake the thread up.
                        // If we didn't set a context, it'll add itself
                        // back to the end of the queue and go to sleep again.
                        worker.notify();
                        worker = null;
                    }
                    
                    // If we didn't set a context, put ourselves to sleep
                    // until we want to check again or get woken up externally.
                    if (context == null) {
                        logger.internal("Threadpool %s sleeping", Threadpool.this.getName());
                        synchronized(this) { this.wait(Threadpool.this.idleTime * 1000); }
                    }
                }
                catch (InterruptedException e) { } 
            }
        }
        
        synchronized void shutdown() {
            this.shutdown = true;
            notifyAll();
        }
    }
    
    private class OperationThread extends Thread {
        private final int id;
        private Operation operation;
        private volatile boolean shutdown;
        
        public OperationThread(int id) {
            this.id = id;
            this.setDaemon(true);
            this.setIdleTitle();
        }
        
        protected void setOperation(Operation operation) {
            assert this.operation == null;
            this.operation = operation;
        }
        
        protected boolean isIdle() {
            return this.operation == null;
        }
        
        private void setIdleTitle() {
            this.setName(Threadpool.this.getName() + this.id + " IDLE");
        }
        
        private void setActiveTitle() {
            this.setName(Threadpool.this.getName() + this.id + " " + this.operation.toString());
        }
        
        @Override
        public synchronized void run() {
            TaskUtils.setThreadEngine(Threadpool.this.engine);
            
            while (!this.shutdown) {
                if (this.operation != null) {
                    Threadpool.this.lastOperationStartTime = new Date();
                    Threadpool.this.operationStartCount.incrementAndGet();
                    this.setActiveTitle();
                    try {
                        this.operation.execute();
                    }
                    // CHECKSTYLE:OFF
                    catch (Throwable t) {
                    // CHECKSTYLE:ON
                        // This should not happen, but if it does, we want to catch it so we don't lose the thread and
                        // we can diagnose later.
                        logger.error(MsgKey.FLOW_OPERATION_EXECUTE_THREAD_EXCEPTION, t, t.getLocalizedMessage());
                    }
                    this.operation = null;
                    Threadpool.this.lastOperationStopTime = new Date();
                    this.setIdleTitle();
                }
                else {
                    try {
                        Threadpool.this.idleThreads.add(this);
                        this.wait();
                    }
                    catch (InterruptedException e) { }
                }
            }
        }

        synchronized void shutdown() {
            this.shutdown = true;
            notifyAll();
        }
    }
}
