/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/tasks/Log.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow.tasks;

import java.util.List;

import com.netapp.common.MsgKey;
import com.netapp.common.flow.Task;
import com.netapp.common.logging.LogLevel;
import com.netapp.common.logging.MessageKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ClassUtils;
import com.netapp.common.util.StringUtils;

/**
 * Log a message outside of another task. The Log message is added to the list of messages
 * generated by the workflow.
 * 
 * The use of this task is similar to the following
 * 
 * <code>
 *   <Log>
 *     <Input name="logLevel">WARN</Input>                             # Text from LogLevel enum
 *     <Input name="messageKeyClass">com.netapp.common.MsgKey</Input>  # full package name of the MsgKey class to use for the log message
 *     <Input name="messageKey">MY_MSG_KEY</Input>                     # Name of the key in the MsgKey class enum to use
 *     <Input name="messageArgs">arg1</Input>
 *     <Input name="messageArgs">arg2</Input>
 *   </Log>
 * </code>
 * 
 * All inputs except for the messageArgs must be supplied.
 * 
 * @author joshb
 *
 */
public class Log implements Task {
    private static final NALogger LOGGER = NALogger.getLogger(Log.class);

    // Task inputs
    protected String logLevel;
    protected String messageKeyClass;
    protected String messageKey;
    protected List<String> messageArgs;

    // internal variables
    protected Class<? extends MessageKey> classForMessageKeyClass;
    
    public void execute() throws Exception {
        if (!validate()) {
            LOGGER.error(MsgKey.FLOW_TASK_LOG_ERROR);
            return;
        }

        MessageKey msgKeyEnum = (MessageKey) ClassUtils.createObject(getClassForMessageKeyClass(), getMessageKey());
        LogLevel level = LogLevel.valueOf(getLogLevel());

        Object[] args = null;

        if (null != getMessageArgs()) {
            args = new Object[getMessageArgs().size()];
            
            for (int i = 0; i < args.length; i++) {
                args[i] = getMessageArgs().get(i);
            }
        }

        LOGGER.log(level, msgKeyEnum, args);
    }

    /**
     * validate the inputs to the task
     * @return <code>true</code> if the inputs to the task are valid
     */
    protected boolean validate() {
        boolean validated = true;

        if (StringUtils.isBlank(getLogLevel())) {
            LOGGER.warn(MsgKey.FLOW_TASK_LOG_ERROR_MISSING_LOG_LEVEL);
            validated = false;
        }

        if (StringUtils.isBlank(getMessageKeyClass())) {
            LOGGER.warn(MsgKey.FLOW_TASK_LOG_ERROR_MISSING_MESSAGE_KEY_CLASS);
            validated = false;
        }

        if (StringUtils.isBlank(getMessageKey())) {
            LOGGER.warn(MsgKey.FLOW_TASK_LOG_ERROR_MISSING_MESSAGE_KEY);
            validated = false;
        }

        try {
            Class<?> clazz = Class.forName(getMessageKeyClass());

            if (!clazz.isEnum()) {
                LOGGER.error(MsgKey.FLOW_TASK_LOG_ERROR_CLASS_NOT_AN_ENUM, getMessageKeyClass());
                validated = false;
            }

            if (!ClassUtils.typeAssignableFrom(MessageKey.class, clazz)) {
                LOGGER.error(MsgKey.FLOW_TASK_LOG_ERROR_NOT_MESSAGE_KEY_CLASS, getMessageKeyClass());
                validated = false;
            }
            setClassForMessageKeyClass(clazz.asSubclass(MessageKey.class));
        }
        catch (ClassNotFoundException e) {
            LOGGER.warn(MsgKey.FLOW_TASK_LOG_ERROR_CANNOT_LOAD_CLASS, e, getMessageKeyClass());
            validated = false;
        }

        return validated;
    }

    /*
     * Task inputs
     */
    protected String getLogLevel() {
        return logLevel;
    }

    public void setLogLevel(String logLevel) {
        this.logLevel = logLevel;
    }

    protected String getMessageKeyClass() {
        return messageKeyClass;
    }

    public void setMessageKeyClass(String messageKeyClass) {
        this.messageKeyClass = messageKeyClass;
    }

    protected String getMessageKey() {
        return messageKey;
    }

    public void setMessageKey(String messageKey) {
        this.messageKey = messageKey;
    }

    protected List<String> getMessageArgs() {
        return messageArgs;
    }

    public void setMessageArgs(List<String> messageArgs) {
        this.messageArgs = messageArgs;
    }

    /*
     * Internal variables
     */

    protected Class<? extends MessageKey> getClassForMessageKeyClass() {
        return classForMessageKeyClass;
    }

    protected void setClassForMessageKeyClass(Class<? extends MessageKey> classForMessageKeyClass) {
        this.classForMessageKeyClass = classForMessageKeyClass;
    }

}
