/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/bootstrap/com/netapp/common/logging/generator/MsgKeyGenerator.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.logging.generator;


import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.regex.Matcher;

import com.netapp.common.Version;
import com.netapp.common.generator.NetappClassGenerator;
import com.netapp.common.generator.NetappCodeFormatter;
import com.netapp.common.logging.Bundle;
import com.netapp.common.logging.LoggingException;
import com.netapp.common.logging.MessageKey;
import com.netapp.common.util.ClassUtils;

/**
 * Construct the MsgKey enumerations for a given bundle file.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public final class MsgKeyGenerator extends NetappClassGenerator {
    private static final String BUNDLE_FILE_SUFFIX = ".properties";
    
    /** Class name of our MsgKey interface */
    private static final String MSG_KEY_INTERFACE = MessageKey.class.getName();

    private static final NetappCodeFormatter FORMATTER;
    static {
        FORMATTER = new NetappCodeFormatter();
        FORMATTER.addHeaderLine("Generated by MsgKeyGenerator " + Version.VERSION + " (" + Version.CHANGEID + ")");
    }
    
    protected String className;
    protected String packageName;
    protected String bundleName;
    protected ResourceBundle bundle;

    protected String classData;

    /**
     * Create a new generator for the supplied class and properties
     * bundle.
     * 
     * @param qualifiedClassName
     * @param bundleName
     * @exception LoggingException An error was detected in the supplied properties bundle.
     */
    public MsgKeyGenerator(String qualifiedClassName, String bundleName) {
        super(ClassUtils.extractPackageName(qualifiedClassName), FORMATTER);
        this.className = ClassUtils.extractUnqualifiedClassName(qualifiedClassName);
        
        if (MSG_KEY_INTERFACE.equals(qualifiedClassName)) {
            throw new LoggingException("Generated class name may not be " + MSG_KEY_INTERFACE);
        }
        
        if (bundleName.endsWith(BUNDLE_FILE_SUFFIX)) {
            bundleName = bundleName.substring(0, bundleName.length() - BUNDLE_FILE_SUFFIX.length());
        }
        this.bundleName = bundleName;
        this.bundle = ResourceBundle.getBundle(bundleName, Locale.US);
        this.generateMsgKeyClass();
    }

    public String getClassData() {
        return this.classData;
    }

    /**
     * Convert a java style camelCase string to an enumeration name
     * Convert to upper case, replacing '.' with '_' and adding a '_' before
     * uppercase letters.
     * @param input
     * @return String converted to Enum format
     */
    static String toEnumString(String input) {
        return input.substring(0, 1).toUpperCase()
            + input.substring(1).replaceAll("\\.", "_").replaceAll("([A-Z])", "_$1").toUpperCase();
    }

    /**
     * Generate the MsgKey enumeration into the current class buffer
     * 
     * @exception LoggerException an error occured processing the input
     */
    private void generateMsgKeyClass() {
        addImport(Bundle.class.getCanonicalName());

        addLine("@Bundle(\"" + bundleName + "\")");

        // Print our class name.
        // We always implement the interface by complete name,
        // as the user may name the resulting class MsgKey as well.
        addLine("public enum " + className + " implements " + MSG_KEY_INTERFACE + " {");
        increaseIndent();

        generateMsgEnumerations();

        // Generate the constructor and our interface implementation
        addLine();
        addLine("private final String key;");
        addLine("private " + className + "(String key) { this.key = key; }");
        addLine("public String getPropertyName() { return this.key; }");
        
        addLine("public static " + className + " parseKey(String key) {");
        increaseIndent();
        addLine("for (" + className + " testKey : " + className + ".values()) {");
        increaseIndent();
        addLine("if (testKey.key.equals(key)) {");
        increaseIndent();
        addLine("return testKey;");
        decreaseIndent();
        addLine("}");
        decreaseIndent();
        addLine("}");
        addLine("throw new IllegalArgumentException(\"No enum found in \" + " + className + ".class.getName() + \" for key \" + key);");
        decreaseIndent();
        addLine("}");
        decreaseIndent();
        
        addLine("}");
    }

    /**
     * 
     */
    private void generateMsgEnumerations() {
        /* Keep a list of message codes and keys we've seen already */
        Set<String> seenCodes = new HashSet<String>();
        Set<String> seenKeys = new HashSet<String>();

        // Iterate over the key message pairs
        Enumeration<String> keys = bundle.getKeys();
        while (keys.hasMoreElements()) {
            String enumVal = keys.nextElement();
            String comment = bundle.getString(enumVal);
            String enumKey = toEnumString(enumVal);

            if (seenKeys.contains(enumKey)) {
                throw new LoggingException("Duplicate enumeration key " + enumKey);
            }
            seenKeys.add(enumKey);

            Matcher matcher = MessageKey.CODE_PATTERN.matcher(comment);
            if (matcher.matches()) {
                String code = matcher.group(1);
                if (seenCodes.contains(code)) {
                    throw new LoggingException("Duplicate message code " + code);
                }
                seenCodes.add(code);
            }

            addJavadocLine(comment);

            addLine(enumKey + "(\"" + enumVal + "\")" + (keys.hasMoreElements() ? "," : ";"));
        }
    }

    public static void main(String[] args) {
        if (args.length < 3) {
            printUsage();
        }

        String propFile = args[0];
        String sourceRoot = args[1];
        String className = args[2];

        try {
            String classFileName = File.separator + ClassUtils.classNameToFilePath(className) + ".java";
            File outputDir = new File(sourceRoot);

            if (!outputDir.exists() || !outputDir.isDirectory()) {
                throw new LoggingException("Output source directory " + outputDir + " does not exist");            
            }

            File classFile = new File(outputDir, classFileName);
            File classDir = classFile.getParentFile();
            if (!classDir.exists() && !classDir.mkdirs()) {
                throw new LoggingException("Unable to create package directory " + classDir);
            }

            MsgKeyGenerator generator = null;

            generator = new MsgKeyGenerator(className, propFile);

            Writer outputWriter = new FileWriter(classFile);
            generator.writeCode(outputWriter);
            outputWriter.flush();
            outputWriter.close();
        }
        catch (LoggingException e) {
            printUsage(e);
        }
        catch (IOException e) {
            printUsage(e);
        }
    }

    private static void printUsage() {
        System.err.println("Usage: MsgKeyGenerator <bundle> <src_root> <class>");
    }

    private static void printUsage(Throwable t) {
        printUsage();
        if (t != null) {
            t.printStackTrace(System.err);
        }
        System.exit(1);
    }
}
