/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/aix/AIXDeviceAnalyzer.java#1 $
 *
 * Copyright (c) 2006 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform.aix;

import java.io.File;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.platform.DeviceAnalyzer;
import com.netapp.common.platform.DeviceType;
import com.netapp.common.platform.PlatformException;
import com.netapp.common.platform.RawDeviceMapping;

/**
 * Provide a simple mapping between block devices and raw on AIX systems.
 * On AIX we only currently support full devices.
 *
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
class AIXDeviceAnalyzer implements DeviceAnalyzer {
    private static final NALogger logger = NALogger.getLogger(AIXDeviceAnalyzer.class);

    /** Pattern to match raw devices. */
    protected static final Pattern RAW_PATTERN = Pattern.compile("(\\/dev\\/)r(hdisk.*)");

    /** Pattern to match block devices. */
    protected static final Pattern BLOCK_PATTERN = Pattern.compile("(\\/dev\\/)(hdisk.*)");

    public AIXDeviceAnalyzer() {
        super();
    }

    /**
     * Map a raw device to the block device. This simply strips the leading r
     * off the device name (/dev/rdev ==&gt; /dev/dev)
     *
     * @param rawDevice The raw device path to map.
     * @return The associate block device
     */
    public RawDeviceMapping lookupMappingByRawDevice(File rawDevice) {
        Matcher m = RAW_PATTERN.matcher(rawDevice.getAbsolutePath());
        if (m.matches()) {
            File disk = new File(m.group(1) + m.group(2));
            return new RawDeviceMapping(rawDevice, disk);
        }
        else {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNSUPPORTED_OPERATION);
            throw new PlatformException(error);
        }
    }

    /**
     * Map a disk to the corresponding raw device.
     *
     * @param disk Disk entry to map
     * @return Raw device name
     */
    public RawDeviceMapping lookupMappingByDisk(File disk) {
        Matcher m = BLOCK_PATTERN.matcher(disk.getAbsolutePath());
        if (m.matches()) {
            File rawDevice = new File(m.group(1) + "r" + m.group(2));
            return new RawDeviceMapping(rawDevice, disk);
        }
        else {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNSUPPORTED_OPERATION);
            throw new PlatformException(error);
        }
    }

    public int extractPartitionNumber(File path) {
        throw new UnsupportedOperationException();
    }
    
    /**
     * AIX doesn't have partitions, if you split a disk it's all done through
     * LVM. Therefore this API doesn't really work.
     */
    public String extractPartitionPathElement(File path) {
        return "";
    }

    /**
     * We only support full disks for now, not LVM LVs.
     * If the supplied string is /dev/hdisk?, we just return that
     * otherwise null.
     * TODO: We should probably add a throws clause to the interface for
     * this to throw exceptions for invalid arguments.
     * 
     * We may eventually need to figure out how
     * to drill into volume groups.
     */
    public String extractFullDiskPathElement(File file) {
        if (BLOCK_PATTERN.matcher(file.getAbsolutePath()).matches()) {
            return file.getAbsolutePath();
        }
        return null;
    }

    public File deviceForPartition(File disk, int partitionNumber) {
        String partition = Integer.toString(partitionNumber);
        return this.deviceForPartition(disk, partition);
    }
    
    public File deviceForPartition(File disk, String partition) {
        return new File(disk + partition);
    }

    public DeviceType determineDeviceType(File file) {
        if (RAW_PATTERN.matcher(file.getAbsolutePath()).matches()) {
            return DeviceType.RAW_DEVICE;
        }
        else if (BLOCK_PATTERN.matcher(file.getAbsolutePath()).matches()) {
            return DeviceType.BASE_DEVICE;
        }
        else {
            return DeviceType.NON_DEVICE;    
        }
    }
}
