/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/unix/OperatingSystemImpl.java#1 $
 * 
 * Copyright (c) 2006 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform.unix;

import java.io.File;
import java.io.IOException;
import java.util.List;

import com.netapp.common.MsgKey;
import com.netapp.common.execution.ShellCommand;
import com.netapp.common.execution.ShellExecutor;
import com.netapp.common.execution.ShellExecutorFactory;
import com.netapp.common.execution.ShellResult;
import com.netapp.common.execution.ShellRunnable;
import com.netapp.common.logging.MessageBundle;
import com.netapp.common.logging.NALogger;
import com.netapp.common.platform.Authenticator;
import com.netapp.common.platform.CredentialNotFoundException;
import com.netapp.common.platform.FileOwner;
import com.netapp.common.platform.FileSecurityDescriptor;
import com.netapp.common.platform.Group;
import com.netapp.common.platform.JNIException;
import com.netapp.common.platform.MountpointMapper;
import com.netapp.common.platform.OperatingSystem;
import com.netapp.common.platform.PlatformException;
import com.netapp.common.platform.User;
import com.netapp.common.platform.FileSecurityDescriptor.AccessControl;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.Utils;

/**
 * Contains functionality shared between most or all Posix compliant
 * operating systems.
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public abstract class OperatingSystemImpl extends OperatingSystem {
    
    private static NALogger logger = NALogger.getLogger();
    
    private FileSystemNative fileSystemNative;
    
    /**
     *
     */
    public OperatingSystemImpl() {
        super();
    }

    /**
     * @see com.netapp.common.platform.OperatingSystem#getPID()
     */
    @Override
    public Long getPID() {
        return getProcessNative().getPID();
    }

    private static final int KILL_RETRIES = 5;

    /**
     * @see com.netapp.common.platform.OperatingSystem#kill(long, int)
     */
    @Override
    public void kill(long pid, int signal) {
        logger.info(MsgKey.PLATFORM_ATTEMPTING_KILL, Utils.numberToStringNoCommas(pid));
        /*
         * Keep sending signals to the process until we get an error from kill().
         * Only then are we sure it is dead.
         */
        boolean isKilled = false;
        int tries = KILL_RETRIES;
        do {
            isKilled = (getProcessNative().kill(pid, signal) == 0);
            if (!isKilled) {
                try {
                    Thread.sleep(300);
                    logger.internal("Retrying kill of PID " + Utils.numberToStringNoCommas(pid));
                }
                catch (InterruptedException e) {
                    logger.internal("Interrupted while sleeping", e);
                }
            }
        }
        while (!isKilled && --tries > 0);

        if (isKilled) {
            logger.info(MsgKey.PLATFORM_KILLED_PROCESS, Utils.numberToStringNoCommas(pid));
        }
        else {
            throw new RuntimeException(logger.error(MsgKey.PLATFORM_CANNOT_KILL,
                new Object[] { pid, KILL_RETRIES }));
        }

        
    }

    @Override
    public boolean isProcessRunning(Long pid) {
        ShellExecutor shellExecutor = ShellExecutorFactory.createShellExecutor();
        ShellResult psResult = shellExecutor.execute(new ShellCommand("ps", "-p", pid.toString()));
        
        /*
         * If ps -p returns 0 then the process is running
         */
        return psResult.isZeroExitCode();
    }

    @Override
    public boolean isUserPrivileged() {
        return System.getProperty("user.name").equals("root");
    }

    @Override
    public File getUserApplicationDirectory() {
        return this.getUserHomeDir();
    }

    @Override
    public FileSecurityDescriptor createFileSecurityDescriptor(User user, Group group, AccessControl... accessControls) {
        FileSecurityDescriptorImpl descriptor = new FileSecurityDescriptorImpl();
        
        FileOwner owner = new FileOwner(user, group);
        
        descriptor.setFileOwner(owner);
        descriptor.addAccessControls(accessControls);
        
        return descriptor;
    }
    
    @Override
    public void changeFileSecurity(File file, FileSecurityDescriptor securityDescriptor) {
        assert file != null;
        assert securityDescriptor != null;
        
        if (!(securityDescriptor instanceof FileSecurityDescriptorImpl)) {
            throw new IllegalArgumentException(logger.debug(MsgKey.PLATFORM_CHANGE_FILE_SECURITY_INCORRECT_TYPE, FileSecurityDescriptorImpl.class.getName(), securityDescriptor.getClass().getName()));
        }
        
        FileSecurityDescriptorImpl securitySettings = (FileSecurityDescriptorImpl)securityDescriptor;
        
        logger.internal("chmod %s %s", securitySettings.getPermissionOctal(), getFilePath(file));
        
        try {
            getFileSystemNative().chmod(file.getAbsolutePath(), securitySettings.getPermissionBits());
        }
        catch (JNIException e) {
            throw new PlatformException(
                logger.debug(MsgKey.PLATFORM_UNIX_CHMOD_FAILED, securitySettings.getPermissionOctal(), getFilePath(file),
                    e.getMessage(), e.getErrno()), e);
        }
        
        if (securityDescriptor.getFileOwner() != null && securityDescriptor.getFileOwner().getUser() != null 
                && securityDescriptor.getFileOwner().getGroup() != null) {
            
            changeOwner(file, securityDescriptor.getFileOwner());
        }
    }

    @Override
    public void changeOwner(File file, FileOwner fileOwner) {
        assert file != null;
        assert fileOwner != null;
        
        if (!(fileOwner.getUser() instanceof UserImpl)) {
            throw new IllegalArgumentException(logger.debug(MsgKey.PLATFORM_CHANGE_FILE_OWNER_INCORRECT_USER_TYPE, UserImpl.class.getName(), (fileOwner.getUser() != null ? fileOwner.getUser().getClass().getName() : "null")));
        }
        
        if (!(fileOwner.getGroup() instanceof GroupImpl)) {
            throw new IllegalArgumentException(logger.debug(MsgKey.PLATFORM_CHANGE_FILE_OWNER_INCORRECT_GROUP_TYPE, GroupImpl.class.getName(), (fileOwner.getGroup() != null ? fileOwner.getGroup().getClass().getName() : "null")));
        }
        
        int uid = ((UserImpl)fileOwner.getUser()).getUid();
        int gid = ((GroupImpl)fileOwner.getGroup()).getGid();
        
        try {
            getFileSystemNative().chown(file.getAbsolutePath(), uid, gid);
        }
        catch (JNIException e) {
            throw new PlatformException(
                logger.debug(MsgKey.PLATFORM_UNIX_CHOWN_FAILED, uid, gid, getFilePath(file), e.getMessage(), e.getErrno()), e);
        }
    }

    @Override
    public boolean canExecute(File executeable, User user) {
        FileSecurityDescriptor descriptor = getFileSecurityDescriptor(executeable);
        
        // ALL_EXECUTE is set
        if (descriptor.getEnabledAccessControls().contains(AccessControl.ALL_EXECUTE)) {
            return true;
        }
        
        // The given user is the owner of the file and OWNWER_EXECUTE is set
        if (descriptor.getEnabledAccessControls().contains(AccessControl.OWNER_EXECUTE)
                && descriptor.getFileOwner().getUser().equals(user)) {
            return true;
        }
        
        // GROUP_EXECUTE is set and the given user is in the group that is the 
        // group owner of the file
        if (descriptor.getEnabledAccessControls().contains(AccessControl.GROUP_EXECUTE)
                && userInGroup(user.getName(), descriptor.getFileOwner().getGroup().getName())) {
            return true;
        }
        return false;
    }

    @Override
    public User getUserFromUsername(String username) {
        assert username != null;
        // Synchronized because getpwnam(3) isn't reentrant
        synchronized (getFileSystemNative()) {
            int uid = getFileSystemNative().getUIDFromUsername(username);
            if (uid < 0) {
                String error = logger.error(MsgKey.PLATFORM_UNIX_USER_NOT_FOUND, username);
                throw new CredentialNotFoundException(error);
            }
            
            return new UserImpl(uid, username);
        }
    }

    @Override
    public Group getGroupFromGroupname(String groupname) {
        assert groupname != null;
        // Synchronized because getgrnam(3) isn't reentrant
        synchronized (getFileSystemNative()) {
            int gid = getFileSystemNative().getGIDFromGroupname(groupname);

            if (gid < 0) {
                String error = logger.error(MsgKey.PLATFORM_UNIX_GROUP_NOT_FOUND, groupname);
                throw new CredentialNotFoundException(error);
            }
            
            return new GroupImpl(gid, groupname);
        }
    }

    @Override
    public MountpointMapper createMountpointMapper() {
        return new UnixMountpointMapper(this);
    }

    public DFResult runDFCommand(List<File> fileList) {
        ShellResult shellResult = ShellExecutorFactory.createShellExecutor().execute(this.createDFCommand(fileList));
        return this.createDFResultConstructor().construct(fileList, shellResult);
    }

    protected abstract ShellRunnable createDFCommand(List<File> fileList);

    protected abstract DFResultConstructor createDFResultConstructor();

    public FileSecurityDescriptor getFileSecurityDescriptor(File file) {
        // Synchronized because it isn't reentrant
        synchronized (getFileSystemNative()) {
            String path;
            try {
                path = file.getCanonicalPath(); 
            }
            catch (IOException e) {
                throw new PlatformException(MessageBundle.localizeMessage(MsgKey.PLATFORM_GET_FILE_SECURITY_DESCRIPTOR_FAILED, file.getAbsolutePath(), e.getMessage()), e);
            }
            
            try {
                FileSecurityDescriptor fileSecurityDescriptor = getFileSystemNative().getFileSecurityDescriptor(path);
                if (fileSecurityDescriptor == null) {
                    throw new PlatformException(MessageBundle.localizeMessage(MsgKey.PLATFORM_GET_FILE_SECURITY_DESCRIPTOR_FAILED, path, "returned null"));
                }
                return fileSecurityDescriptor;
            }
            catch (JNIException e) {
                throw new PlatformException(MessageBundle.localizeMessage(MsgKey.PLATFORM_GET_FILE_SECURITY_DESCRIPTOR_FAILED, path, e.getMessage()), e);
            }
        }
    }
    
    protected synchronized FileSystemNative getFileSystemNative() {
        if (this.fileSystemNative == null) {
            try {
                Class<?> nativeClass = Class.forName("com.netapp.common.platform.unix.FileSystemNativeImpl");
                this.fileSystemNative = (FileSystemNative)nativeClass.newInstance();
            }
            catch (UnsatisfiedLinkError e) {
                logger.warn(MsgKey.PLATFORM_PROCESS_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
            catch (ClassNotFoundException e) {
                logger.warn(MsgKey.PLATFORM_PROCESS_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
            catch (InstantiationException e) {
                logger.warn(MsgKey.PLATFORM_PROCESS_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
            catch (IllegalAccessException e) {
                logger.warn(MsgKey.PLATFORM_PROCESS_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
        }
        
        if (this.fileSystemNative == null) {
            throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_FILE_SYSTEM_METHOD_DEPENDS_ON_NATIVE_LIBRARY));
        }
        
        return this.fileSystemNative;
    }
    
    @Override
    public boolean userInGroup(String username, String groupname) {
        return getFileSystemNative().userInGroup(username, groupname);
    }
    
    /**
     * @see com.netapp.common.platform.OperatingSystem#createAuthenticator(ConfigBundle)
     */
    @Override
    public Authenticator createAuthenticator(ConfigBundle configBundle) {
        return new UnixAuthenticatorNativeImpl(configBundle);
    }

    /**
     * It is unclear how this will be handled for several of the UNIX systems.
     * Currently, we know we can retrieve the UUID on Linux and probably FreeBSD
     * even when running under VMware. We are unsure how to perform this on the
     * other platforms at this time.
     * 
     */
    public String getBiosUuid() {
        throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_OS_UUID_NOT_IMPLEMENTED));
    }
}
