/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/windows/OperatingSystemImpl.java#1 $
 *
 * Copyright (c) 2006 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform.windows;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import com.netapp.common.MsgKey;
import com.netapp.common.execution.ShellCommand;
import com.netapp.common.execution.ShellExecutorFactory;
import com.netapp.common.logging.LogLevel;
import com.netapp.common.logging.MessageBundle;
import com.netapp.common.logging.MessageKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.platform.Authenticator;
import com.netapp.common.platform.CredentialNotFoundException;
import com.netapp.common.platform.DeviceAnalyzer;
import com.netapp.common.platform.DeviceType;
import com.netapp.common.platform.FileOwner;
import com.netapp.common.platform.FileSecurityDescriptor;
import com.netapp.common.platform.FileSystemMount;
import com.netapp.common.platform.Group;
import com.netapp.common.platform.JNIException;
import com.netapp.common.platform.MountpointMapper;
import com.netapp.common.platform.OperatingSystem;
import com.netapp.common.platform.PartitionTable;
import com.netapp.common.platform.PlatformException;
import com.netapp.common.platform.RawDeviceMapping;
import com.netapp.common.platform.User;
import com.netapp.common.platform.FileSecurityDescriptor.AccessControl;
import com.netapp.common.platform.windows.cimv2.Win32ComputerSystemProduct;
import com.netapp.common.platform.windows.cimv2.Win32Directory;
import com.netapp.common.platform.windows.cimv2.Win32Volume;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.ConfigProperty;
import com.netapp.common.util.StringUtils;


public class OperatingSystemImpl extends OperatingSystem {
    
    private static NALogger logger = NALogger.getLogger();
    /**
     * Values of "os.arch" that should be considered x86
     */
    protected static final ConfigProperty<String> VERSIONS_WINDOWS_ARCH_X86 = ConfigProperty.stringProperty("versions.windows.arch.x86");
   
    /**
     * Values of "os.arch" that should be considered x64
     */
    protected static final ConfigProperty<String> VERSIONS_WINDOWS_ARCH_X64 = ConfigProperty.stringProperty("versions.windows.arch.x64");
 
    private static final String WMI_CIMV2_NAMESPACE = "root\\cimv2";
    
    private WindowsRegistry windowsRegistry;

    private FileSystemNative fileSystemNative;
    
    private SystemEventNative systemEventNative;
    
    @Override
    public String getArch() {
        String jvmArch = System.getProperty("os.arch");
        String x86 = VERSIONS_WINDOWS_ARCH_X86.lookup(this.getPlatformConfigBundle());
        String x64 = VERSIONS_WINDOWS_ARCH_X64.lookup(this.getPlatformConfigBundle());
        
        if (jvmArch.matches(x86)) {
            return "x86";
        }
        else if (jvmArch.matches(x64)) {
            return "x64";
        }
        return jvmArch;
    }
    
    public WindowsRegistry getWindowsRegistry() {
        if (this.windowsRegistry == null) {
            this.windowsRegistry = new WindowsRegistry();
        }
        
        return this.windowsRegistry;
    }
    
    @Override
    public File getUserApplicationDirectory() {
        return new File(this.getUserHomeDir(), "Application Data");
    }

    @Override
    public DeviceAnalyzer createDeviceAnalyzer() {
        /*
         * TODO
         * Break this out into a separate class
         */
        return new DeviceAnalyzer() {

            public DeviceType determineDeviceType(File file) {
                if (file.getPath().matches("\\\\\\\\.\\\\PHYSICALDRIVE\\d+")) {
                    return DeviceType.BASE_DEVICE;
                }
                return DeviceType.NON_DEVICE;
                // throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_NO_IMPLEMENTATION));
            }

            public File deviceForPartition(File disk, int partitionNumber) {
                String partition = Integer.toString(partitionNumber);
                return this.deviceForPartition(disk, partition);
            }
            public File deviceForPartition(File disk, String partition) {
                throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_NO_IMPLEMENTATION));
            }

            public String extractFullDiskPathElement(File file) {
                throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_NO_IMPLEMENTATION));
            }

            public int extractPartitionNumber(File file) {
                throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_NO_IMPLEMENTATION));
            }

            public String extractPartitionPathElement(File file) {
                throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_NO_IMPLEMENTATION));
            }

            public RawDeviceMapping lookupMappingByDisk(File disk) {
                throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_NO_IMPLEMENTATION));
            }

            public RawDeviceMapping lookupMappingByRawDevice(File rawDevice) {
                throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_NO_IMPLEMENTATION));
            }
        };
    }
    
    @Override
    public Collection<FileSystemMount> getMountedFilesystems() {
        WMIConnection wmi = WMIConnection.getConnection(WMI_CIMV2_NAMESPACE);
        Collection<Win32Volume> volumes = wmi.query(Win32Volume.class, "SELECT * FROM Win32_Volume");
        
        List<FileSystemMount> mounts = new ArrayList<FileSystemMount>();
        for (Win32Volume volume : volumes) {
            Collection<Win32Directory> mountPoints = wmi.associatorsOf(Win32Directory.class, volume, "Win32_Directory");
            if (mountPoints != null) {
                for (Win32Directory mountPoint : mountPoints) {
                    mounts.add(new FileSystemMount(null, volume.getDeviceID(), mountPoint.getName(), volume.getFileSystem(), ""));
                }
            }
        }
        return mounts;
    }
        
    @Override
    public String getLibraryEnvironmentName() {
        /*
         * Returns a "known" environment variable that the *.cmd scripts can then take and append to "-Djava.library.path ..." 
         */
        return "APP_LIBRARY_PATH";
    }

    @Override
    public Long getPID() {
        return new Long(getProcessNative().getPID());
    }

    @Override
    public boolean isProcessRunning(Long pid) {
        return getProcessNative().isPIDAlive(pid.intValue()) == 1;
    }

    @Override
    public boolean isUserPrivileged() {
        // TODO: implement this method for Windows
        return true;
    }

    @Override
    public void kill(long pid, int signal) {
        getProcessNative().kill((int) pid, signal);
    }

    /**
     * The currently implementation on Windows will modify the owner and primary group of the file
     * and explicitly set permissions for the file owner, the file group, and the Windows "Everyone"
     * group to enable or disable read, write, and execute corresponding to the settings of the security
     * descriptor permissions.
     * 
     * Because we currently utilize the GetEffectiveRights API on Windows when getting the rights of a file
     * it is possible for a caller to attempt to remove write permission for the owner but because of the group 
     * permission or some other granted rights that the owner still effectively have write permission.
     * 
     * @param file The file to change the security on
     * @param securityDescriptor The owner and permissions to set on the file
     */
    @Override
    public void changeFileSecurity(File file, FileSecurityDescriptor securityDescriptor) {
        assert file != null;
        assert securityDescriptor != null;
        
        if (!(securityDescriptor instanceof FileSecurityDescriptorImpl)) {
            throw new IllegalArgumentException(logger.debug(MsgKey.PLATFORM_CHANGE_FILE_SECURITY_INCORRECT_TYPE, FileSecurityDescriptorImpl.class.getName(), securityDescriptor.getClass().getName()));
        }
        FileSecurityDescriptorImpl securitySettings = (FileSecurityDescriptorImpl)securityDescriptor;
        
        // If for some reason we got a descriptor with no owner, read the current settings
        // Who's fracking idea was it to add this useless FileOwner object in here instead
        // of just putting User and Group on the Descriptor itself?
        FileOwner newOwner = securitySettings.getFileOwner();
        if (newOwner == null 
                || newOwner.getUser() == null
                || newOwner.getGroup() == null) {
            FileSecurityDescriptor old = this.getFileSecurityDescriptor(file);
            
            User newUser = (newOwner == null || newOwner.getUser() == null) ? old.getFileOwner().getUser() : newOwner.getUser();
            Group newGroup = (newOwner == null || newOwner.getGroup() == null) ? old.getFileOwner().getGroup() : newOwner.getGroup();
            
            securitySettings.setFileOwner(new FileOwner(newUser, newGroup));
        }
        
        logger.internal("chmod %s %s", securitySettings, getFilePath(file)); 
        try {
            getFileSystemNative().setFileSecurityDescriptor(file.getAbsolutePath(), securitySettings);
        }
        catch (JNIException e) {
            throw new PlatformException(
                logger.debug(MsgKey.PLATFORM_WINDOWS_CHANGE_PERMISSIONS_FAILED, getFilePath(file), 
                    securitySettings.getPermissionOctal(), e.getErrno(), e.getMessage()), e);
        }
        
    }

    @Override
    public void changeOwner(File file, FileOwner fileOwner) {
        assert file != null;
        assert fileOwner != null;
        
        if (!(fileOwner.getUser() instanceof UserImpl)) {
            throw new IllegalArgumentException(logger.debug(MsgKey.PLATFORM_CHANGE_FILE_OWNER_INCORRECT_USER_TYPE, UserImpl.class.getName(), (fileOwner.getUser() != null ? fileOwner.getUser().getClass().getName() : "null")));
        }
        
        if (!(fileOwner.getGroup() instanceof GroupImpl)) {
            throw new IllegalArgumentException(logger.debug(MsgKey.PLATFORM_CHANGE_FILE_OWNER_INCORRECT_GROUP_TYPE, GroupImpl.class.getName(), (fileOwner.getGroup() != null ? fileOwner.getGroup().getClass().getName() : "null")));
        }

        logger.internal("chown %s:%s %s", fileOwner.getUser(), fileOwner.getGroup(), getFilePath(file));
        try {
            getFileSystemNative().changeOwner(file.getAbsolutePath(), fileOwner);
        }
        catch (JNIException e) {
            throw new PlatformException(
                logger.debug(MsgKey.PLATFORM_WINDOWS_CHANGE_OWNER_FAILED,
                    file.getAbsolutePath(), fileOwner.getUser(), fileOwner.getGroup(), e.getErrno(), e.getMessage()), e);
        }
    }

    @Override
    public Group getGroupFromGroupname(String groupname) {
        try {
            return getFileSystemNative().getGroupFromGroupname(groupname);
        }
        catch (JNIException e) {
            throw new CredentialNotFoundException(
                logger.error(MsgKey.PLATFORM_WINDOWS_GROUP_NOT_FOUND, groupname, e.getMessage()), e);
        }
    }

    @Override
    public User getUserFromUsername(String username) {
        try {
            return getFileSystemNative().getUserFromUsername(username);
        }
        catch (JNIException e) {
            throw new CredentialNotFoundException(
                logger.error(MsgKey.PLATFORM_WINDOWS_USER_NOT_FOUND, username, e.getMessage()), e);
        }
    }

    @Override
    public FileSecurityDescriptor getFileSecurityDescriptor(File file) {
        try {
            String path = file.getCanonicalPath();
            if (path == null) {
                // This is a severe error condition - used to be an assert
                throw new PlatformException(MessageBundle.localizeMessage(
                    MsgKey.PLATFORM_GET_FILE_SECURITY_DESCRIPTOR_FAILED, path, "no canonical path"));
            }
            FileSecurityDescriptor descriptor = getFileSystemNative().getFileSecurityDescriptor(path);

            // This SHOULD NOT HAPPEN anymore - Any event that would return null
            // should throw a JNIException
            if (descriptor == null) {
                throw new PlatformException(MessageBundle.localizeMessage(
                    MsgKey.PLATFORM_GET_FILE_SECURITY_DESCRIPTOR_FAILED, path, "returned null"));
            }
            return descriptor;
        }
        catch (IOException e) {
            throw new PlatformException(MessageBundle.localizeMessage(
                MsgKey.PLATFORM_GET_FILE_SECURITY_DESCRIPTOR_FAILED, file.getAbsolutePath(), e.getMessage()), e);
        }
        catch (JNIException e) {
            throw new PlatformException(MessageBundle.localizeMessage(
                MsgKey.PLATFORM_GET_FILE_SECURITY_DESCRIPTOR_FAILED, file.getAbsolutePath(), e.getMessage()), e);
        }
    }

    @Override
    public boolean canExecute(File executeable, User user) {
        return getFileSystemNative().canExecute(executeable.getAbsolutePath(), (UserImpl)user);
    }
    
    @Override
    public FileSecurityDescriptor createFileSecurityDescriptor(User user, Group group, AccessControl... accessControls) {
        FileSecurityDescriptorImpl descriptor = new FileSecurityDescriptorImpl();
        
        FileOwner owner = null;
        
        if (user != null || group != null) {
            owner = new FileOwner(user, group);
        }
        
        descriptor.setFileOwner(owner);
        descriptor.addAccessControls(accessControls);
        
        return descriptor;
    }

    @Override
    public boolean isWindows() {
        return true;
    }

    @Override
    public MountpointMapper createMountpointMapper() {
        return new WindowsMountpointMapper();
    }

    public ServiceState getServiceState(String serviceName) {
        ServiceControlResult serviceControlResult = new ServiceControlResult(
            ShellExecutorFactory.createShellExecutor().execute(
                new ShellCommand("sc", "query", serviceName)), serviceName);

        return serviceControlResult.getServiceState();
    }
    
    public String getWindowsRoot() {
        return System.getenv("SystemRoot");
    }

    @Override
    public PartitionTable readPartitionTable(File device) {
        throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_NO_IMPLEMENTATION));
    }

    protected synchronized FileSystemNative getFileSystemNative() {
        if (this.fileSystemNative == null) {
            try {
                Class<?> nativeClass = Class.forName("com.netapp.common.platform.windows.FileSystemNativeImpl");
                this.fileSystemNative = (FileSystemNative)nativeClass.newInstance();
            }
            catch (UnsatisfiedLinkError e) {
                logger.warn(MsgKey.PLATFORM_PROCESS_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
            catch (ClassNotFoundException e) {
                logger.warn(MsgKey.PLATFORM_PROCESS_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
            catch (InstantiationException e) {
                logger.warn(MsgKey.PLATFORM_PROCESS_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
            catch (IllegalAccessException e) {
                logger.warn(MsgKey.PLATFORM_PROCESS_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
        }
        
        if (this.fileSystemNative == null) {
            throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_FILE_SYSTEM_METHOD_DEPENDS_ON_NATIVE_LIBRARY));
        }
        
        return this.fileSystemNative;
    }
    
    public Authenticator createAuthenticator(ConfigBundle configBundle) {
        return new WindowsAuthenticatorNativeImpl(configBundle);
    }
    
    /**
     * For Windows, we can get the BIOS UUID by retrieving the UUID from the Win32_ComputerSystemProduct object in WMI.
     * The first three segments come back with the hex pairs reversed, so we have to do a slight translation on the
     * string coming back so that it will match the value that the outside callers might expect.
     */
    public String getBiosUuid() {
        WMIConnection wmi = WMIConnection.getConnection(WMI_CIMV2_NAMESPACE);
        Collection<Win32ComputerSystemProduct> values = wmi.query(Win32ComputerSystemProduct.class, "SELECT * FROM Win32_ComputerSystemProduct");
        
        if (1 != values.size()) {
            throw new PlatformException(logger.error(MsgKey.PLATFORM_WINDOWS_WMI_MULTIPLE_COMPUTER_SYSTEM_PRODUCTS, values.size()));
        }
        
        String biosUuid = null;
        
        for (Win32ComputerSystemProduct product : values) {
            biosUuid = product.getUuid();
        }
        
        if (!StringUtils.isBlank(biosUuid)) {
            String[] parts = biosUuid.split("-");
            StringBuilder builder = new StringBuilder();
            
            parts[0] = reverseHex(parts[0]);
            parts[1] = reverseHex(parts[1]);
            parts[2] = reverseHex(parts[2]);
            
            for (String part : parts) {
                if (0 != builder.length()) {
                    builder.append("-");
                }
                
                builder.append(part);
            }
            
            biosUuid = builder.toString();
        }
        

        return biosUuid;
    }
    
    /**
     * Reverse the string of hex values.
     * 
     * @param input
     * @return
     */
    protected String reverseHex(String input) {
        StringBuilder builder = new StringBuilder();
        
        for (int i = 0; i < input.length(); i += 2) {
            builder.insert(0, input.subSequence(i, i + 2));
        }
        
        return builder.toString();
    }
    
    protected synchronized SystemEventNative getSystemEventNative() {
        if (this.systemEventNative == null) {
            try {
                Class<?> nativeClass = Class.forName("com.netapp.common.platform.windows.SystemEventNativeImpl");
                this.systemEventNative = nativeClass.asSubclass(SystemEventNative.class).newInstance();
            }
            catch (UnsatisfiedLinkError e) {
                logger.debug(MsgKey.PLATFORM_SYSTEM_EVENT_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
            catch (ClassNotFoundException e) {
                logger.debug(MsgKey.PLATFORM_SYSTEM_EVENT_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
            catch (InstantiationException e) {
                logger.debug(MsgKey.PLATFORM_SYSTEM_EVENT_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
            catch (IllegalAccessException e) {
                logger.debug(MsgKey.PLATFORM_SYSTEM_EVENT_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
        }
        return this.systemEventNative;
    }
    
    protected void emitSystemEvent(LogLevel level, String source, String message) {
        SystemEventNative se = getSystemEventNative();
        if (se != null) {
            se.emitSystemLog(level, source, message);
        }
    }
    
    @Override
    public void systemEvent(LogLevel level, String source, MessageKey key) {
        emitSystemEvent(level, source, MessageBundle.localizeMessage(key));
    }
    
    @Override
    public void systemEvent(LogLevel level, String source, MessageKey key, Throwable t) {
        emitSystemEvent(level, source, MessageBundle.localizeMessage(key));
    }
    
    @Override
    public void systemEvent(LogLevel level, String source, MessageKey key, Object... params) {
        emitSystemEvent(level, source, MessageBundle.localizeMessage(key, params));
    }
    
    @Override
    public void systemEvent(LogLevel level, String source, MessageKey key, Throwable t, Object... params) {
        emitSystemEvent(level, source, MessageBundle.localizeMessage(key, params));
    }
    
    @Override
    public boolean userInGroup(String username, String groupname) {
        throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_NO_IMPLEMENTATION));
    }

    @Override
    public String toString() {
        return MessageBundle.localizeMessage(MsgKey.PLATFORM_WINDOWS_DESCRIPTION,
            getPlatformBrand(), getArch());
    }

    /*
    public static void main(String [] args) {
        OperatingSystem os = OperatingSystem.createInstance();
        System.out.println("BIOS UUID = " + os.getBiosUuid());
    }
    */
}
