/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/windows/WindowsRegistry.java#1 $
 * 
 * Copyright (c) 2006 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform.windows;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.CollectionUtils;

/**
 * Provides a proxy for the Windows Registry.  Accessed via:
 *     com.netapp.common.platform.windows.OperatingSystem.getWindowsRegistry()
 * 
 * @author grimshaw
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class WindowsRegistry {
    private static final NALogger logger = NALogger.getLogger(WindowsRegistry.class);
    
    private RegistryNative registryNative;
    
    /**
     * Look up the information from the Windows Registry.
     * 
     * For example: key: HKEY_LOCAL_MACHINE\SOFTWARE\Network Appliance\SnapDrive
     *              itemName: InstallPath
     *              
     * Result: C:\Program Files\NetApp\SnapDrive\
     * 
     * @return String The value found in the Windows Registry.
     */
    public String getRegistryValue(String key, String itemName) {
        String result = null;
            
        Map<String, String> keyValues = getKeyItemValues(key);
            
        if (keyValues.containsKey(itemName)) {
            // This is a valid SubKey 
            result = keyValues.get(itemName);
        }
        
        return result;
    }
    
    /**
     * For example: key: HKEY_LOCAL_MACHINE\SOFTWARE\ORACLE
     * 
     * Result: HKEY_LOCAL_MACHINE\SOFTWARE\ORACLE\KEY_OraDb10g_home1
     *         HKEY_LOCAL_MACHINE\SOFTWARE\ORACLE\SYSMAN
     *         
     * @param key
     * @return List of subkeys under key
     */
    public List<String> getSubKeys(String key) {
        List<String> result = new ArrayList<String>();
        
        StringTokenizer st = runRegistryQuery(key);
        
        while (st.hasMoreTokens()) {
            String next = st.nextToken().trim();
            
            if (next.startsWith(key) && !next.equals(key)) {
                result.add(next);
            }
        }
        
        return result;
    }
    
    /** 
     * For example: key: HKEY_LOCAL_MACHINE\SOFTWARE\Network Appliance\SnapDrive
     *              
     * Result: InstallPath --> C:\Program Files\NetApp\SnapDrive\
     *         ProductVersion --> 5.0.0.1886
     *         UpgradeState --> 0
     */
    public Map<String, String> getKeyItemValues(String key) {
        Map<String, String> result = new HashMap<String, String>();
        
        StringTokenizer st = runRegistryQuery(key);
        
        boolean include = false;
        
        while (st.hasMoreTokens()) {
            String nextLine = st.nextToken().trim();
                      
            if (nextLine.equals(key)) {
                include = true;
                continue;
            }
                                    
            if (include) {
                if (nextLine.equals("")) {
                    // Blank line means end of key values.
                    break;
                }       
                StringTokenizer stKeyValues = null;
                
                if (nextLine.contains("\t")) {
                    stKeyValues = new StringTokenizer(nextLine, "\t");
                }
                else {
                    stKeyValues = new StringTokenizer(nextLine, " "); 
                } 
                String subKeyName  = stKeyValues.nextToken().trim();  
                String subKeyType  = stKeyValues.nextToken().trim();  
                
                String subKeyValue = "";
                if (stKeyValues.hasMoreTokens()) {
                    subKeyValue = stKeyValues.nextToken().trim();
                }
                
                while (stKeyValues.hasMoreTokens()) {
                    // The value might be something like 'C:\Program Files\fdsdfssd' so handle
                    //   the case where the tokenizer splits the 'Program' from the 'Files'.
                    subKeyValue += " " + stKeyValues.nextToken().trim();
                }
     
                result.put(subKeyName, subKeyValue);
            }   
        }
        
        return result;
    } 
    
    protected StringTokenizer runRegistryQuery(String key) {
        StringTokenizer result = null;

        List<String> queryResult = getRegistryNative().queryRegistry(key);
        String resultString = CollectionUtils.collectionToString(queryResult, "\r\n"); // reg.exe outputs \r\n, which we are mimicing        
        result = new StringTokenizer(resultString, "\n");
                                        
        return result;
    }

    protected synchronized RegistryNative getRegistryNative() {
        if (this.registryNative == null) {
            try {
                Class<?> registryNativeClass = Class.forName("com.netapp.common.platform.windows.NativeRegistryImpl");
                this.registryNative = (RegistryNative)registryNativeClass.newInstance();
            }
            catch (UnsatisfiedLinkError e) {
                logger.warn(MsgKey.PLATFORM_REGISTRY_UNABLE_TO_LOAD_NATIVE);
            }
            catch (ClassNotFoundException e) {
                logger.warn(MsgKey.PLATFORM_REGISTRY_UNABLE_TO_LOAD_NATIVE);
            }
            catch (InstantiationException e) {
                logger.warn(MsgKey.PLATFORM_REGISTRY_UNABLE_TO_LOAD_NATIVE);
            }
            catch (IllegalAccessException e) {
                logger.warn(MsgKey.PLATFORM_REGISTRY_UNABLE_TO_LOAD_NATIVE);
            }
        }
        
        if (this.registryNative == null) {
            throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_REGISTRY_METHOD_DEPENDS_ON_NATIVE_LIBRARY));
        }
        
        return this.registryNative;
    }

}
