/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/remote/RemoteObjectFactoryImpl.java#1 $
 * 
 * Copyright (c) 2008 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.remote;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.rmi.server.Unreferenced;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.MessageBundle;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ClassUtils;
import com.netapp.common.util.GUID;

public class RemoteObjectFactoryImpl extends UnicastRemoteObject implements RemoteObjectFactory, Unreferenced {

    private static final NALogger logger = NALogger.getLogger();
    private GUID guid;
    private String registryHost;
    private int registryPort;
    private Registry remoteObjectFactoryRegistry;
    
    public RemoteObjectFactoryImpl(String registryHost, int registryPort, GUID guid) throws RemoteException {
        this.registryHost = registryHost;
        this.registryPort = registryPort;
        this.guid = guid;
        this.bind();
    }

    private void bind() throws RemoteException {

        logger.internal("RemoteObjectFactoryImpl finding registry at %s:%s", this.registryHost, this.registryPort);
        try {
            this.remoteObjectFactoryRegistry = RemoteObjectFactoryRegistryImpl.getRegistry(this.registryHost, this.registryPort);
            logger.internal("RemoteObjectFactoryImpl binding to %s", this.getBindURL());
            this.remoteObjectFactoryRegistry.rebind(this.guid.toString(), this);
            logger.internal("RemoteObjectFactoryImpl bound to %s", this.getBindURL());
        }
        catch (RemoteException e) {
            logger.internal("RemoteObjectFactoryImpl bind error.", e);
            throw e;
        }
    }
    
    private void unbind() throws RemoteException {
        try {
            this.remoteObjectFactoryRegistry.unbind(this.guid.toString());
        }
        catch (NotBoundException e) {
            // Ignore
        }
    }

    public <T extends Remote> T newRemoteObject(Class<T> clazz, Object... constructorArgs) throws RemoteException {
        try {
            Class<?>[] constructorClassArgs = ClassUtils.extractClasses(constructorArgs);
            Constructor<T> constructor = clazz.getConstructor(constructorClassArgs);
            constructor.setAccessible(true);
            T object = constructor.newInstance(constructorArgs);
            if (object instanceof RemoteObjectFactoryAware) {
                ((RemoteObjectFactoryAware) object).setDestroyRunnable(new Runnable() {
                    public void run() {
                        try {
                            RemoteObjectFactoryImpl.this.destroy();
                        }
                        catch (RemoteException e) {
                            // Ignore
                        }
                    }
                });
            }
            return object;
        }
        catch (InstantiationException e) {
            throw new RuntimeException(e);
        }
        catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        }
        catch (SecurityException e) {
            throw new RuntimeException(e);
        }
        catch (NoSuchMethodException e) {
            throw new RuntimeException(e);
        }
        catch (IllegalArgumentException e) {
            throw new RuntimeException(e);
        }
        catch (InvocationTargetException e) {
            throw new RuntimeException(e);
        }
    }

    public void destroy() throws RemoteException {
        this.unbind();
        
        /*
         * Wait 1 second to allow the RMI call to return successfully, then exit this JVM
         */
        ScheduledExecutorService scheduledExecutorService = Executors.newSingleThreadScheduledExecutor();
        scheduledExecutorService.schedule(new Runnable() {
            public void run() {
                System.out.println(MessageBundle.localizeMessage(MsgKey.REMOTE_REMOTE_OBJECT_FACTORY_SHUTDOWN));
                System.exit(0);
            }
        }, 1, TimeUnit.SECONDS);
    }

    public void unreferenced() {
        logger.internal("RemoteObjectFactory unreferenced");
        System.exit(0);
    }

    public void ping() throws RemoteException {
        // No-op
    }
    
    String getBindURL() {
        return String.format("%s:%s/%s", this.registryHost, this.registryPort, this.guid);
    }
    
    public static void main(String[] args) throws RemoteException {
        if (args.length != 3) {
            throw new RuntimeException("Usage: " + RemoteObjectFactoryImpl.class.getName() + " <rmi registry host> <rmi registry port> <guid>");
        }
        
        String registryHost = args[0];
        String registryPort = args[1];
        GUID guid = new GUID(args[2]);

        RemoteObjectFactoryImpl remoteObjectFactoryImpl = new RemoteObjectFactoryImpl(registryHost, Integer.valueOf(registryPort), guid);
        logger.info(MsgKey.REMOTE_REMOTE_OBJECT_FACTORY_STARTED, remoteObjectFactoryImpl.getBindURL());
    }

}
