/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/util/ConfigProperty.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import java.io.File;
import java.util.Properties;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.MessageBundle;
import com.netapp.common.logging.NALogger;

/**
 * Typed configuration property used to lookup items from an application
 * {@link ConfigBundle}.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class ConfigProperty<T> {
    private static final NALogger logger = NALogger.getLogger(ConfigProperty.class);
    
    protected final String property;
    protected final Class<T> type;
    protected final T defaultValue;
    protected boolean isRequired;
    
    /**
     * Construct a new property of the specified type and property name
     * @param type
     * @param property
     */
    public ConfigProperty(Class<T> type, String property) {
        this(type, property, null);
    }

    /**
     * Construct a new property of the specified type, property name, and default value.
     * @param type
     * @param property
     * @param defaultValue
     */
    public ConfigProperty(Class<T> type, String property, T defaultValue) {
        this.type = type;
        this.property = property;
        this.defaultValue = defaultValue;
    }
    
    public String getProperty() {
        return this.property;
    }
    
    public Class<T> getType() {
        return this.type;
    }
    
    public T getDefaultValue() {
        return this.defaultValue;
    }

    /**
     * Mark this ConfigProperty as required.
     * Causes an exception to be thrown upon {@link ConfigProperty#lookup(ConfigBundle)} if the property is not found.  
     * 
     * @return A reference to the ConfigProperty (the this object).  Useful for method chaining / fluent-style invocation.
     */
    public ConfigProperty<T> required() {
        if (this.defaultValue != null) {
            throw new IllegalStateException("Cannot mark a ConfigProperty with a default value as required");
        }
        this.isRequired = true;
        return this;
    }
    
    /**
     * Retrieve the value of this property from the supplied bundle
     * @param bundle
     * @return value of this property
     */
    public T lookup(ConfigBundle bundle) {
        return lookup(bundle.getProperties());
    }
    
    /**
     * Retrieve the value of this property from the supplied properties.
     * @param the properties
     * @return value of this property
     */
    public T lookup(Properties properties) {
        T value = this.defaultValue;
        
        String propValue = properties.getProperty(this.property);
        if (propValue != null) {
            try {
                value = ClassUtils.createObject(this.type, propValue);    
            }
            // CHECKSTYLE:OFF
            catch (Exception e) {
            // CHECKSTYLE:ON
                logger.error(MsgKey.UTIL_CONFIG_EXCEPTION_LOADING_PROPERTY, e, this.type, this.property, propValue, e.getLocalizedMessage());
            }
        }
        else {
            if (this.isRequired) {
                throw new RuntimeException(MessageBundle.localizeMessage(MsgKey.UTIL_CONFIG_PROPERTY_IS_NULL, this.property));
            }
            else {
                logger.debug(MsgKey.UTIL_CONFIG_PROPERTY_IS_NULL, this.property);    
            }
        }

        return value;
    }
    
    /**
     * Convenience function to construct a new String property.
     * @param property Property file key
     * @return String property for the specified key
     */
    public static ConfigProperty<String> stringProperty(String property) {
        return new ConfigProperty<String>(String.class, property);
    }
    
    public static ConfigProperty<String> stringProperty(String property, String defaultValue) {
        return new ConfigProperty<String>(String.class, property, defaultValue);
    }
    
    /**
     * Convenience function to construct a new boolean property.
     * @param property Property file key
     * @return boolean property for the specified key
     */
    public static ConfigProperty<Boolean> booleanProperty(String property) {
        return new ConfigProperty<Boolean>(Boolean.class, property);
    }
    
    public static ConfigProperty<Boolean> booleanProperty(String property, Boolean defaultValue) {
        return new ConfigProperty<Boolean>(Boolean.class, property, defaultValue);
    }
    
    /**
     * Convenience function to construct a new integer property.
     * @param property Property file key
     * @return integer property for the specified key
     */
    public static ConfigProperty<Integer> intProperty(String property) {
        return new ConfigProperty<Integer>(Integer.class, property);
    }
    
    /**
     * Convenience function to construct a new integer property.
     * @param property Property file key
     * @param defaultValue default value if not specified
     * @return integer property for the specified key
     */
    public static ConfigProperty<Integer> intProperty(String property, Integer defaultValue) {
        return new ConfigProperty<Integer>(Integer.class, property, defaultValue);
    }
    
    /**
     * Convenience function to construct a new long property.
     * @param property Property file key
     * @return long property for the specified key
     */
    public static ConfigProperty<Long> longProperty(String property) {
        return new ConfigProperty<Long>(Long.class, property);
    }
    
    /**
     * Convenience function to construct a new long property.
     * @param property Property file key
     * @param defaultValue default value
     * @return long property for the specified key
     */
    public static ConfigProperty<Long> longProperty(String property, Long defaultValue) {
        return new ConfigProperty<Long>(Long.class, property, defaultValue);
    }
    
    /**
     * Convenience function to construct a new double property.
     * @param property Property file key
     * @return double property for the specified key
     */
    public static ConfigProperty<Double> doubleProperty(String property) {
        return new ConfigProperty<Double>(Double.class, property);
    }
    
    /**
     * Convenience function to construct a new double property.
     * @param property Property file key
     * @param defaultValue default value
     * @return double property for the specified key
     */
    public static ConfigProperty<Double> doubleProperty(String property, Double defaultValue) {
        return new ConfigProperty<Double>(Double.class, property, defaultValue);
    }

    /**
     * Convenience function to construct a new float property.
     * @param property Property file key
     * @return float property for the specified key
     */
    public static ConfigProperty<Float> floatProperty(String property) {
        return new ConfigProperty<Float>(Float.class, property);
    }
    
    /**
     * Convenience function to construct a new float property.
     * @param property Property file key
     * @param defaultValue default value
     * @return float property for the specified key
     */
    public static ConfigProperty<Float> floatProperty(String property, Float defaultValue) {
        return new ConfigProperty<Float>(Float.class, property, defaultValue);
    }
    
    /**
     * Convenience function to construct a new file property.
     * @param property Property file key
     * @return file property for the specified key
     */
    public static ConfigProperty<File> fileProperty(String property) {
        return new FileProperty(property);
    }
    
    /**
     * Convenience function to construct a new file property.
     * @param property Property file key
     * @param defaultValue defaultValue default value
     * @return file property for the specified key
     */
    public static ConfigProperty<File> fileProperty(String property, File defaultValue) {
        return new FileProperty(property, defaultValue);
    }
    
    /**
     * Convenience function to construct a new Interval property.
     * @param property Property file key
     * @return Interval property for the specified key
     */
    public static ConfigProperty<Interval> intervalProperty(String property) {
        return new ConfigProperty<Interval>(Interval.class, property);
    }
    
    /**
     * Convenience function to construct a new Interval property.
     * @param property Property file key
     * @param defaultValue default value
     * @return Interval property for the specified key
     */
    public static ConfigProperty<Interval> intervalProperty(String property, Interval defaultValue) {
        return new ConfigProperty<Interval>(Interval.class, property, defaultValue);
    }
    
    /**
     * Convenience function to construct a new Interval property.
     * @param property Property file key
     * @param defaultValue default value
     * @return Interval property for the specified key
     */
    public static ConfigProperty<Interval> intervalProperty(String property, String defaultValue) {
        return new ConfigProperty<Interval>(Interval.class, property, new Interval(defaultValue));
    }
    
    @Override
    public String toString() {
        return this.property;
    }
}
