/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/util/FileUtils.java#1 $
 * 
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.netapp.common.logging.NALogger;

public class FileUtils {
    private static final NALogger logger = NALogger.getLogger(FileUtils.class);
    
    private FileUtils() {
    }

    /**
     * Waits the specified number of milliseconds for a file to appear
     * (when file.exists() returns true).
     * 
     * @param file
     * @param timeoutInMillis
     * @return true if the file exists
     */
    public static boolean waitForFileToAppear(File file, long timeoutInMillis) {
        if (!file.exists()) {
            long startTime = System.currentTimeMillis();
            while (!file.exists() && (System.currentTimeMillis() - startTime) < timeoutInMillis) {
                try {
                    Thread.sleep(100L);
                }
                catch (InterruptedException e) {
                    break;
                }
            }
        }
        return file.exists();
    }

    /**
     * Print a line to a file. Convinience method for quick IO operations.
     * @param content
     * @param file File to print contents to. If file already exists, it will
     * be truncated and replaced with content.
     */
    public static void echo(File file, String... content) throws IOException {
        echo(file, false, content);
    }
    
    /**
     * Print a line to a file. Convinience method for quick IO operations.
     *
     * @param file
     * @param append
     * @param content
     */
    public static void echo(File file, boolean append, String... content) throws IOException {
        PrintStream stream = new PrintStream(new FileOutputStream(file, append));
        for (String line : content) {
            stream.println(line);
        }
        stream.close();
    }
    
    /**
     * Create a temp file in the system temp directory with the given contents
     * @param prefix
     * @param content
     * @throws IOException
     */
    public static File createTempFile(String prefix, String... content) throws IOException {
        File file = File.createTempFile(prefix, "tmp");
        file.deleteOnExit();
        echo(file, content);
        return file;
    }
    
    /**
     * grep a file for a regular expression, returning the first
     * Matcher that matched.
     *
     * @param pattern line pattern to search for
     * @param file file to search
     * @return Matcher of the first line that matched, or null if
     * no matches were found.
     */
    public static Matcher matchLine(Pattern pattern, File file) throws IOException {
        List<Matcher> result = matchLines(pattern, file, 1);
        if (!result.isEmpty()) {
            return result.get(0);
        }
        return null;
    }

    /**
     * grep a file for a regular expression, returning any Matchers
     * that matched.
     *
     * @param pattern line pattern to search for
     * @param file file to search
     * @return List of Matchers that matched. If no lines matched, an empty list
     * will be returned
     */
    public static List<Matcher> matchLines(Pattern pattern, File file) throws IOException {
        return matchLines(pattern, file, 0);
    }

    /**
     * grep a file for a regular expression, returning any Matchers
     * that matched.
     *
     * @param pattern line pattern to search for
     * @param file file to search
     * @param Maximum number of results to find. If 0, will match the entire file.
     * @return List of Matchers that matched. If no lines matched, an empty list
     * will be returned
     */
    public static List<Matcher> matchLines(Pattern pattern, File file, int maxResults) throws IOException {

        BufferedReader in = new BufferedReader(new FileReader(file));
        List<Matcher> results = new ArrayList<Matcher>();

        String line = in.readLine();
        while (line != null) {
            Matcher m = pattern.matcher(line);
            if (m.matches()) {
                results.add(m);
                if (results.size() == maxResults) {
                    break;
                }
            }
            line = in.readLine();
        }
        in.close();
        return results;
    }
    
    /**
     * Returns the canonical path of the file, but
     * if that fails, returns the absolute path of the file.
     * This differs from {@link File#getCanonicalFile()} in that
     * this method does not throw IOException.
     * 
     * @param file File for which to get the path
     * @return the canonical path of the file, but
     * if that fails, returns the absolute path of the file.
     */
    public static File getCanonicalFile(File file) {
        file = file.getAbsoluteFile();
        try {
            file = file.getCanonicalFile();
        }
        catch (IOException e) { }
        return file;
    }
}
