/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/ApiData.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import com.netapp.common.MsgKey;
import com.netapp.common.zapi.annotation.Api;
import com.netapp.common.zapi.annotation.Input;
import com.netapp.common.zapi.annotation.Optional;
import com.netapp.common.zapi.annotation.Output;

/**
 * Represents the reflective properties of a zapi {@Api} call.
 */
final class ApiData {
    /** ApiData cache */
    private static Map<Class<?>, ApiData> apis = new ConcurrentHashMap<Class<?>, ApiData>();
    
    /** The java class of this api */
    protected Class<?> apiClass;
    
    /** The zapi call name */
    protected String zapiName;
    
    /** Input arguments to the ZAPI call. Key is the java field name */
    protected Map<String, FieldData> inputFields;
    
    /** Output from the ZAPI call. Key is the zapi element name */
    protected Map<String, FieldData> outputFields;
    
    /**
     * Construct a new ApiData object for the specified class annotated
     * with the {@link Api} annotation.
     * @param apiClass Class representing a zapi call
     */
    private ApiData(Class<?> apiClass) {
        this.apiClass = apiClass;
        Api api = this.apiClass.getAnnotation(Api.class);
        if (api == null) {
            throw new ZAPIRunnerException(
                ZAPIRunner.logger.error(MsgKey.ZAPI_RUNNER_CLASS_NOT_ZAPI_REQUEST, this.apiClass.getName()));
        }
        
        if (api.value().length() > 0) {
            this.zapiName = api.value();
        }
        else {
            this.zapiName = ZAPIRunner.javaToZapi(apiClass.getSimpleName());
        }
        
        inputFields = new HashMap<String, FieldData>();
        outputFields = new HashMap<String, FieldData>();
        
        // Check for annotations on the getters and setters
        Method[] methods = this.apiClass.getDeclaredMethods();
        for (Method m : methods) {
            boolean optional = m.isAnnotationPresent(Optional.class);
            Input input = m.getAnnotation(Input.class);
            if (input != null) {
                FieldData fieldData = new FieldData(m, input.value(), optional);
                if (!this.inputFields.containsKey(fieldData.getJavaFieldName())) {
                    this.inputFields.put(fieldData.getJavaFieldName(), fieldData);
                }
                else {
                    throw new ZAPIRunnerException(ZAPIRunner.logger.error(MsgKey.ZAPI_RUNNER_INPUT_FIELD_DECLARED_TWICE, m, this.apiClass));
                }
            }
            
            Output output = m.getAnnotation(Output.class);
            if (output != null) {
                FieldData fieldData = new FieldData(m, output.value(), optional);
                if (!this.outputFields.containsKey(fieldData.getZapiFieldName())) {
                    this.outputFields.put(fieldData.getZapiFieldName(), fieldData);
                }
                else {
                    throw new ZAPIRunnerException(ZAPIRunner.logger.error(
                        MsgKey.ZAPI_RUNNER_OUTPUT_FIELD_DECLARED_TWICE, 
                        fieldData.getZapiFieldName(), 
                        this.apiClass));
                }
            }
        }
    }
    
    /**
     * Retrieve the ApiData structure for a specific api call instance.
     * This is a convenience function identical to {@link #getApiData(Class) getApiData(obj.getClass())}
     * 
     * @param obj Instantiated Api call to lookup
     * @return ApiData for the specified api call
     */
    public static ApiData getApiData(Object obj) {
        return getApiData(obj.getClass());
    }
    
    /**
     * Retrieve the ApiData structure for a specific api class.
     * @param clazz Class of an api object to lookup
     * @return ApiData for the specified api call
     */
    public static ApiData getApiData(Class<?> clazz) {
        ApiData apiData = apis.get(clazz);
        if (apiData == null) {
            apiData = new ApiData(clazz);
            apis.put(clazz, apiData);
        }
        return apiData;
    }
}