/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/FieldData.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi;

import static com.netapp.common.util.ClassUtils.*;
import static com.netapp.common.util.StringUtils.*;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Type;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;

/**
 * Represents the reflexive properties of a zapi {@com.netapp.common.zapi.annotation.Input}, {@com.netapp.common.zapi.annotation.Output}, or {@com.netapp.common.zapi.annotation.Element}
 * mapped to a Java field.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
final class FieldData {
    private static final NALogger logger = NALogger.getLogger(FieldData.class);
    /** 
     * The java generic type of this field and Collection interface (if defined as a zapi array)
     */
    private Type type;
    
    /** The java class of the {@link com.netapp.common.zapi.annotation.Api} or {@com.netapp.common.zapi.annotation.Typedef} declaring this class */
    private Class<?> javaClass;
    
    /** The java field name */
    private String javaFieldName;
    
    /** The original zapi type of this field */
    private String zapiType;
    
    /** The zapi name of this field */
    private String zapiFieldName;
    
    /** Whether or not this field is optional or required when the enclosing object is populated */
    private boolean optional;
    
    /**
     * Construct a new FieldData from a reflective Method
     * @param method Javabean getter or setter for this field.
     * @param zapiFieldName Declared zapi name of this element. If not specified, the 
     * name will be determined from the java field name
     * @param optional Mark this field as optional or not
     */
    protected FieldData(Method method, String zapiFieldName, boolean optional) {
        this.javaClass = method.getDeclaringClass();
        this.javaFieldName = extractFieldNameFromBeanMethod(method);
        this.type = extractFieldTypeFromBeanMethod(method);
        this.optional = optional;
        if (zapiFieldName.length() > 0) {
            this.zapiFieldName = zapiFieldName;
        }
        else {
            this.zapiFieldName = ZAPIRunner.javaToZapi(this.javaFieldName);
        }
        
        com.netapp.common.zapi.annotation.Type typeAnnotation = method.getAnnotation(com.netapp.common.zapi.annotation.Type.class);
        if (typeAnnotation != null) {
            this.zapiType = typeAnnotation.value();
        }
        validate();
    }
    
    /**
     * Validate this field is of a legal type.
     * 
     * @throws ZAPIRunnerException - the type of this field is not permitted
     *
     */
    private void validate() {
        if (this.getObjectClass().isPrimitive()) {
            throw new ZAPIRunnerException(logger.error(
                MsgKey.ZAPI_RUNNER_PRIMITIVE_NOT_ALLOWED, 
                this.javaClass.getName(), 
                this.javaFieldName, 
                this.type));
        }
    }
    
    /** 
     * Get the java field name of this element 
     * @return java field name for this element
     */
    public String getJavaFieldName() {
        return this.javaFieldName;
    }
    
    /**
     * Get the java type this element represents
     * @return Java generic type for this element
     */
    public Type getType() {
        return this.type;
    }
    
    public String getZapiType() {
        return this.zapiType;
    }
    /**
     * Get the zapi name for this element
     * @return zapi name
     */
    public String getZapiFieldName() {
        return this.zapiFieldName;
    }
    
    /**
     * Check if this field is optional or not
     * @return true if this field is not required to be set
     */
    public boolean isOptional() {
        return this.optional;
    }
    
    /**
     * Check if this element is a collection of items
     * @return true if this field is a collection type
     */
    public boolean isCollection() {
        return isGenericCollection(this.type);
    }
    
    /**
     * Get the Collection class of this field.
     * @return Class of the collection type, or null if this field is not a collection.
     */
    public Class<?> getCollectionClass() {
        return extractCollectionClass(this.type);
    }
    
    /**
     * Get the immediate class of this field. If this field is not a collection,
     * the immediate class is the class of the element itself. If this field is a
     * collection, the immediate class is the same as {@link #getCollectionClass()}
     * @return immediate class of this field
     */
    public Class<?> getImmediateClass() {
        return extractImmediateClass(this.type);
    }
    
    /**
     * Get the ultimate object class of this field. If this field is a collection,
     * the object class is the type contained inside the collection, otherwise this
     * is the same as {@link #getImmediateClass()}
     * @return ultimate object class of this field
     */
    public Class<?> getObjectClass() {
        return extractObjectClass(this.type);
    }
    
    /**
     * Retrieve the value of this field from an object, first by attempting to
     * call getFieldName, then by reading the field directly.
     * @param container Object containing the field to read
     * @return Value of this field on container
     */
    public Object getFieldValue(Object container) {
        Exception reflectException = null;
        Method getter = null;
        
        String regex = "(is|get)" + initialUpperCase(this.javaFieldName);
        Object fieldValue = null;
        for (Method m : container.getClass().getDeclaredMethods()) {
            if (m.getName().matches(regex) && m.getParameterTypes().length == 0) {
                getter = m;
                break;
            }
        }
        
        if (getter != null) {
            try {
                logger.internalTrace("Using getter %s for field %s#%s", getter.getName(), container.getClass().getName(), this.javaFieldName);
                fieldValue = getter.invoke(container);
                return fieldValue;
            }
            catch (IllegalAccessException e) {
                reflectException = e;
            }
            catch (IllegalArgumentException e) {
                reflectException = e;
            }
            catch (InvocationTargetException e) {
                reflectException = e;
            }
        }
        
        if (reflectException != null) {
            throw new ZAPIRunnerException(logger.error(
                MsgKey.ZAPI_RUNNER_READ_REFLECTION_EXCEPTION, 
                reflectException, 
                this.javaFieldName, 
                container.getClass()),
                reflectException);
        }
        
        throw new ZAPIRunnerException(logger.error(
            MsgKey.ZAPI_RUNNER_NO_READ_ACCESS_TO_FIELD, 
            this.javaFieldName, 
            container.getClass()));      
        
    }
    
    /**
     * Attempt to set the value of this field on a given object.
     * @param container Object containing the field to set
     * @param fieldValue Value to set
     */
    public void setFieldValue(Object container, Object fieldValue) {
        Exception reflectException = null;
        Method setter = null;
        
        String regex = "set" + initialUpperCase(this.javaFieldName);
        for (Method m : container.getClass().getDeclaredMethods()) {
            if (m.getName().matches(regex) 
                    && m.getParameterTypes().length == 1
                    && (fieldValue == null || m.getParameterTypes()[0].isInstance(fieldValue))) {
                setter = m;
                break;
            }
        }
        
        if (setter != null) {
            logger.internalTrace("Using setter %s for field %s#%s", setter.getName(), container.getClass().getName(), this.javaFieldName);
            try {
                setter.invoke(container, fieldValue);
                return;
            }
            catch (IllegalArgumentException e) {
                reflectException = e;
            }
            catch (IllegalAccessException e) {
                reflectException = e;
            }
            catch (InvocationTargetException e) {
                reflectException = e;
            }
        }
        
        if (reflectException != null) {
            throw new ZAPIRunnerException(logger.error(
                MsgKey.ZAPI_RUNNER_WRITE_REFLECTION_EXCEPTION, 
                this.javaFieldName, 
                this.getType()));
        }
        throw new ZAPIRunnerException(logger.error(
            MsgKey.ZAPI_RUNNER_NO_WRITE_ACCESS_TO_FIELD,
            this.javaFieldName,
            container.getClass().getName()));
    }
}