/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/ZAPIRunner.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi;

import static com.netapp.common.util.StringUtils.initialUpperCase;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import netapp.manage.NaElement;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.LogLevel;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.StringUtils;
import com.netapp.common.zapi.annotation.Typedef;

/**
 * Run a ZAPI call via annotated classes
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class ZAPIRunner {
    static final NALogger logger = NALogger.getLogger(ZAPIRunner.class);

    protected static final int EAPIERROR = 13001;
    protected static final int EAPINOTFOUND = 13005;
    
    protected ZAPIExecutor executor;
    protected LogLevel invalidListContentLogLevel = LogLevel.DEBUG;
    private Set<String> validApis;
    
    /**
     * Create a new runner
     * @param executor
     */
    public ZAPIRunner(ZAPIExecutor executor) {
        this.executor = executor;
    }
    
    /**
     * Create a ZAPI command and execute
     * @param zapi
     * @throws ZAPIExecuteException 
     */
    public <T> T run(T zapi) throws ZAPIExecuteException {
        ZAPICommand command = buildCommand(zapi);
        logger.trace(MsgKey.ZAPI_RUNNER_ZAPI_REQUEST, zapi.getClass().getName());
        ZAPIResult result = this.executor.execute(command);
        // It's possible for a ZAPIExecutor to return null if there was
        // an error it wants to ignore, but didn't generate any output.
        if (result != null && result.getZapiResponse() != null) {
            parseOutput(result, zapi);
        }
        return zapi;
    }

    public static String zapiToJava(String zapi) {
        StringBuffer b = new StringBuffer();
        String[] words = zapi.split("-");
        for (String word : words) {
            b.append(initialUpperCase(word));
        }
        return b.toString();
    }

    public static String javaToZapi(String java) {
        return java.replaceAll("(?:([a-z])([A-Z0-9])|([A-Z])([0-9]))", "$1-$2").toLowerCase();
    }

    protected static ZAPICommand buildCommand(Object zapi) {
        ApiData apiData = ApiData.getApiData(zapi);
        ZAPICommand command = new ZAPICommand(apiData.zapiName);
        
        Collection<FieldData> fields = apiData.inputFields.values();
        for (FieldData field : fields) {
            NaElement param = buildFieldElement(zapi, field);
            if (param != null) {
                command.addParameter(param);
            }
            else if (!field.isOptional()) {
                throw new ZAPIRunnerException(logger.error(
                    MsgKey.ZAPI_RUNNER_REQUIRED_INPUT_NOT_SET, 
                    field.getZapiFieldName(), 
                    apiData.zapiName)); 
            }
        }
        return command;
    }
    

    
    protected static NaElement buildFieldElement(Object container, FieldData field) {
        Object fieldValue = field.getFieldValue(container);

        if (fieldValue == null) {
            return null;
        }
        
        NaElement element = new NaElement(field.getZapiFieldName());
        if (fieldValue instanceof Collection) {
            // Iterate through the items in the collection
            // Add a child element for each
            Class<?> clazz = field.getObjectClass();
            if (clazz.isAnnotationPresent(Typedef.class)) {
                for (Object child : (Collection<?>)fieldValue) {
                    element.addChildElem(buildTypedefElement(child));
                }
            }
            else {
                String elementName = field.getZapiType().replaceAll("\\[\\]", "");
                for (Object child : (Collection<?>)fieldValue) {
                    NaElement e2 = new NaElement(elementName);
                    e2.setContent(child.toString());
                    element.addChildElem(e2);
                }
            }
        }
        else if (fieldValue.getClass().isAnnotationPresent(Typedef.class)) {
            NaElement child = buildTypedefElement(fieldValue);
            // If the field type and name are the same, they only get packed once
            if (child.getName().equals(field.getZapiFieldName())) {
                element = child;
            }
            else {
                element.addChildElem(buildTypedefElement(fieldValue));
            }
        }
        else if (fieldValue instanceof Date) {
            // Dates are always returned as the number of seconds from epoch.
            long seconds = ((Date)fieldValue).getTime() / 1000;
            element.setContent(Long.toString(seconds));
        }
        else {
            element.setContent(fieldValue.toString());
        }
        return element;
    }

    protected static NaElement buildTypedefElement(Object typedef) {
        TypedefData typedefData = TypedefData.getTypedefData(typedef);
        NaElement element = new NaElement(typedefData.zapiTypedefName);
        
        for (FieldData field : typedefData.zapiElements.values()) {
            NaElement child = buildFieldElement(typedef, field);
            if (child != null) {
                element.addChildElem(child);
            }
            else if (!field.isOptional()) {
                throw new ZAPIRunnerException(logger.error(
                    MsgKey.ZAPI_RUNNER_REQUIRED_INPUT_NOT_SET, 
                    field.getZapiFieldName(), 
                    typedefData.zapiTypedefName)); 
            }
        }
        return element;
    }
    
    protected void parseOutput(ZAPIResult result, Object zapi) {
        parseOutput(result.getZapiResponse(), zapi);
    }
    
    protected void parseOutput(NaElement result, Object zapi) {
        Map<String, FieldData> outputs = ApiData.getApiData(zapi).outputFields;
        fillObject(zapi, outputs, result);
    }
    
    protected void fillObject(Object container, Map<String, FieldData> fieldMap, NaElement element) {
        for (Object o : element.getChildren()) {
            NaElement child = (NaElement)o;
            String outputName = child.getName();

            FieldData fieldData = fieldMap.get(outputName);
            // responses that begin with "is-" are booleans.
            // See if we can find the same field without the is- prefix
            if (fieldData == null && outputName.startsWith("is-")) {
                outputName = outputName.replaceAll("^is-", "");
                fieldData = fieldMap.get(outputName);
            }
            
            if (fieldData == null) {
                logger.debug(MsgKey.ZAPI_RUNNER_OUTPUT_FIELD_NOT_FOUND, outputName, element.getName());
                continue;
            }
            logger.internalTrace("Mapping %s to %s (%s)", child.getName(), fieldData.getJavaFieldName(), fieldData.getType());
            // TODO: This needs some work
            Object fieldValue = null;
            if (fieldData.isCollection()) {
                logger.internalTrace("Field %s is Collection (%s)", outputName, fieldData.getObjectClass());
                fieldValue = createCollectionFromElement(fieldData, child);
            }
            else if (fieldData.getObjectClass().isAnnotationPresent(Typedef.class)) {
                TypedefData fieldTypedef = TypedefData.getTypedefData(fieldData.getObjectClass());
                logger.internalTrace("Field %s is typedef of %s", outputName, fieldTypedef.getTypedefClass());
                if (outputName.equals(fieldTypedef.getZapiTypedefName())) {
                    logger.internalTrace("Field %s is singly packed", outputName);
                    fieldValue = createTypedefObject(fieldTypedef, child);
                }
                else {
                    logger.internal("Field %s is doubly packed %s", outputName, fieldTypedef.getZapiTypedefName());
                    NaElement contents = child.getChildByName(fieldTypedef.getZapiTypedefName());
                    if (contents != null) {
                        fieldValue = createTypedefObject(fieldTypedef, contents);
                    }
                }
            }
            else {
                logger.internalTrace("Field %s is simple type", outputName, fieldData.getObjectClass());
                fieldValue = createSimpleObject(fieldData.getObjectClass(), child.getContent());
            }
            fieldData.setFieldValue(container, fieldValue);
        }
    }
    
    protected static Object createSimpleObject(Class<?> clazz, String content) {
        try {
            if (clazz.equals(String.class)) {
                return content;
            }

            // at this point, if the return happened to be an empty tag (<tag/>)
            // then we will have a content which parsed as null or ""
            // which will generally fail with NumberFormatException or the like for these
            // other data types.  At this point, we should simply set that simple object  
            // to null, since for non string types, that is effectively what we have been told
            if (StringUtils.isEmpty(content)) {
                return null;
            }
            
            if (clazz.equals(Integer.class)) {
                return Integer.parseInt(content);
            }
            
            if (clazz.equals(Long.class)) {
                return Long.parseLong(content);
            }

            if (clazz.equals(BigInteger.class)) {
                return new BigInteger(content);
            }
            
            if (clazz.equals(Boolean.class)) {
                return Boolean.parseBoolean(content);
            }
            
            // Dates are always returned as the number of seconds from epoch.
            // Date(Long) takes milliseconds since epoch
            if (clazz.equals(Date.class)) {
                long l = Long.parseLong(content) * 1000;
                return new Date(l);
            }
        }
        catch (NumberFormatException e) {
            throw new ZAPIRunnerException(logger.error(MsgKey.ZAPI_RUNNER_CANNOT_PARSE_VALUE, e, content, clazz.getSimpleName()), e);
        }
        
        throw new ZAPIRunnerException(logger.error(MsgKey.ZAPI_RUNNER_NO_CONVERSION_AVAILABLE, clazz.getSimpleName()));
    }
    
    /**
     * Create a collection of objects from the list declared in an NaElement.
     * @param fieldData FieldData structure containing the type of class to be created for each element
     * @param element Element representing the list. Each child is a new class to be put in the list
     * @return A Collection<fieldData.getObjectClass()> populated by the children in element
     */
    protected Collection<?> createCollectionFromElement(FieldData fieldData, NaElement element) {
        // TODO: Implement this correctly based on fieldData.getCollectionClass()
        // Right now, everything is a List.
        Collection<Object> col = new ArrayList<Object>();
        
        TypedefData typedef = null;
        String innerFieldName = null;
        // Check if this is complex content or not.
        // We're assuming anything that doesn't have an annotation we can
        // fill in as a simple object with createSimpleObject.
        if (fieldData.getObjectClass().isAnnotationPresent(Typedef.class)) {
            typedef = TypedefData.getTypedefData(fieldData.getObjectClass());
            innerFieldName = typedef.getZapiTypedefName();
        }
        else {
            // If it's simple, the inner element name is the type without the array specifier.
            innerFieldName = fieldData.getZapiType().replace("[]", "");
        }
        
        for (Object o : element.getChildren()) {
            NaElement child = (NaElement)o;
            if (!innerFieldName.equals(child.getName())) {
                // Check the inner element name is what we expect, either from the
                // inner typedef or the simple field name. If it's not, skip and
                // continue. This is usually caused by incorrectly formatted output
                // from the server.
                logger.log(
                    this.invalidListContentLogLevel,
                    MsgKey.ZAPI_RUNNER_COLLECTION_ELEMENT_NOT_EXPECTED_TYPE, 
                    child.getName(), 
                    element.getName(), 
                    innerFieldName);
            }
            else {
                // Create the new object either by recursing into the typedef
                Object obj = null;
                if (typedef != null) {
                    obj = createTypedefObject(typedef, child);
                }
                // or just creating a simple object
                else {
                    obj = createSimpleObject(fieldData.getObjectClass(), child.getContent());
                }
                col.add(obj);
            }
        }
        return col;
    }
    
    /**
     * Create a single object from an NaElement and populate its fields.
     * @param typedefData TypedefData describing the desired class type
     * @param element Element containing the contents to be loaded into the new object
     * @return new object loaded with the values from element
     */
    public Object createTypedefObject(TypedefData typedefData, NaElement element) {
        Object obj = null;
        Exception reflectException = null;
        try {
            obj = typedefData.typedefClass.newInstance();
        }
        catch (InstantiationException e) {
            reflectException = e;
        }
        catch (IllegalAccessException e) {
            reflectException = e;
        }
        if (reflectException != null) {
            throw new ZAPIRunnerException(logger.error(
                MsgKey.ZAPI_RUNNER_INSTANTIATE_REFLECTION_EXCEPTION,
                reflectException,
                typedefData.typedefClass.getName(), 
                typedefData.getZapiTypedefName()), reflectException);
        }
        fillObject(obj, typedefData.getZapiElements(), element);
        return obj;
    }
    
    /**
     * Determine if the API defined in the class zapi is supported by the current target.
     * Results are cached at first invocation.
     * This only works on &gt;= ONTAP 7.2.
     * 
     * @param zapi
     * @return true if the target supports the requested zapi
     * @throws ZAPIRunnerException An error occured executing the request
     * @throws UnsupportedOperationException The target does not implement the system-api-list method
     */
    public synchronized boolean apiAvailable(Class<?> zapi) {
        ApiData api = ApiData.getApiData(zapi);
        if (this.validApis == null) {
            // This is why we don't do this by hand
            try {
                ZAPIResult result = this.executor.execute(new ZAPICommand("system-api-list"));
                this.validApis = new TreeSet<String>();
                NaElement apis = result.getZapiResponse().getChildByName("apis");
                for (Object obj : apis.getChildren()) {
                    NaElement element = (NaElement) obj;
                    if ("system-api-info".equals(element.getName())) {
                        this.validApis.add(element.getChildContent("name"));
                    }
                }
            }
            catch (ZAPIExecuteException e) {
                // Some (all?) versions return EAPIERROR instead of EAPINOTFOUND
                if (e.getErrno() == EAPIERROR || e.getErrno() == EAPINOTFOUND) {
                    throw new UnsupportedOperationException(e);
                }
                throw new ZAPIRunnerException(e.getLocalizedMessage(), e);
            }
        }
        return this.validApis.contains(api.zapiName);
    }
}
