/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/generator/autozapi/AutoZapiFieldGenerator.java#1 $
 * 
 * Copyright (c) 2009 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi.generator.autozapi;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ClassUtils;
import com.netapp.common.util.StringUtils;
import com.netapp.common.zapi.generator.FieldGenerator;
import com.netapp.common.zapi.generator.GeneratorUtils;
import com.netapp.common.zapi.parser.Field;
import com.netapp.common.zapi.parser.Type;

public class AutoZapiFieldGenerator<F extends Field> extends FieldGenerator<F> {
    private static final NALogger logger = NALogger.getLogger(AutoZapiFieldGenerator.class);
    
    
    /**
     * List of base ZAPI types to Java Types.
     * Types that aren't in this list need to be resolved to a Typedef
     */
    static final Map<String, String> SIMPLE_TYPE_MAP = new HashMap<String, String>();  
    static {
        // Types allowed by spec
        SIMPLE_TYPE_MAP.put("string", "java.lang.String");
        SIMPLE_TYPE_MAP.put("boolean", "java.lang.Boolean");
        SIMPLE_TYPE_MAP.put("integer", "java.lang.Long");
        
        // DFM
        SIMPLE_TYPE_MAP.put("bool", "java.lang.Boolean");
        
        // TODO: FIX - A typedef of empty needs a container for attributes
        SIMPLE_TYPE_MAP.put("empty", "java.lang.Object");
    }
    
    public AutoZapiFieldGenerator(AutoZapiGeneratorFactory generatorFactory, F field) {
        super(generatorFactory, field);
    }
    
    @Override
    public String getNameGenerated() {
        if (getTarget().getName() == null) {
            return null;
        }
        
        // Convert the zapi name into a java equivilent.
        // If the field name starts with is-, strip it off
        String tmpName = getTarget().getName().replaceAll("^is-", "").toLowerCase();
        
        tmpName = GeneratorUtils.zapiToLowerCamelCase(tmpName);
        
        // If the field name conflicts with a java type, prepend 'zapi' to avoid
        // a compile id
        if (ClassUtils.isJavaReservedWord(tmpName)) {
            tmpName = "zapi" + StringUtils.initialUpperCase(tmpName);
        }
        
        return tmpName;
    }
    
    /**
     * Get the fully qualified java type for this field, or null
     * if the type is not a recognized simple type.
     * @return Fully qualified java class name for this type
     */
    private String getBuiltinType() {
        String javaType = SIMPLE_TYPE_MAP.get(getTarget().getType());
        
        if (javaType != null && javaType.equals("java.lang.Long")) {
            // Fields that end with -time or -timestamp are Dates
            if (getTarget().getName().endsWith("-time") || getTarget().getName().endsWith("-timestamp")) {
                javaType = "java.util.Date";
            }
            
            // Otherwise if we have a range, use the appropriate integer type
            else if (getTarget().getRange() == null) {
                logger.info(MsgKey.ZAPI_GENERATOR_FIELD_SHOULD_HAVE_RANGE, getTarget().getName(), getTarget().getDeclaringType().getName());
            }
            else if (getTarget().getRange().getBitCount() > Long.SIZE) {
                javaType = "java.math.BigInteger";
            }
            else if (getTarget().getRange().getBitCount() <= Integer.SIZE) {
                javaType = "java.lang.Integer";
            }
            // Otherwise we're just Long
        }
        return javaType;
    }

    /**
     * Get the complete java type for this field.
     * TODO: Move the logging to a validate message
     * This gets called more than once for each field
     * @return String representing the qualified Java Type of this field
     */
    public String getQualifiedTypeClass() {
        String javaType = getBuiltinType();
        if (javaType == null) {
            Type t = getTarget().getTypedefType();
            if (t == null) {
                logger.error(MsgKey.ZAPI_GENERATOR_FIELD_TYPE_NOT_FOUND, getTarget().getType(), getTarget().getName(), getTarget().getDeclaringType());
                javaType = "java.lang.Object";
            }
            else {
                javaType = ((AutoZapiTypeGenerator<? extends Type>) getGeneratorFactory().getTypeGenerator(t)).getQualifiedName();
            }
        }
        return javaType;
    }
    
    public String getUnqualifiedTypeClass() {
        return ClassUtils.extractUnqualifiedClassName(getQualifiedTypeClass());
    }
    
    public String getTypeDeclaration() {
        String typeName = getUnqualifiedTypeClass();
        if (getTarget().isCollection()) {
            typeName = "List<" + typeName + ">";
        }
        return typeName;
    }
    
    /**
     * Get the method suffix for this field.
     * @return method suffix
     */
    public String getMethodSuffix() {
        return StringUtils.initialUpperCase(getNameGenerated());
    }
    
    /**
     * Get the set of imports required to declare this field, including any
     * annotations or types used.
     * @return imports to include for this field
     * @see Type#getImports()
     */
    public Set<String> getImports() {
        Set<String> imports = new HashSet<String>();
        imports.add(AutoZapiTypeGenerator.ANNOTATION_PACKAGE + ".Type");
        
        if (getTarget().isCollection()) {
            imports.add("java.util.List");
        }
        
        if (getTarget().getRange() != null) {
            imports.add(AutoZapiTypeGenerator.ANNOTATION_PACKAGE + ".Range");
        }
        if (getTarget().isOptional()) {
            imports.add(AutoZapiTypeGenerator.ANNOTATION_PACKAGE + ".Optional");
        }
        
        imports.add(getQualifiedTypeClass());
        return imports;
    }

    /**
     * Get the list of annotations to declare on this field.
     * @return List of annotations for this field
     */
    public List<String> getFieldAnnotations() {
        List<String> res = new ArrayList<String>();
        res.add("Type(\"" + getTarget().getType() + (getTarget().isCollection() ? "[]" : "") + "\")");
        if (getTarget().getRange() != null) {
            res.add("Range(min = \"" + getTarget().getRange().getMin() + "\", max = \"" + getTarget().getRange().getMax() + "\")");
        }
        if (getTarget().isOptional()) {
            res.add("Optional");
        }
        return res;

    }

    /**
     * Convert a camel-cased Java field name to ZAPI format
     * @param java java field name
     * @return zapi field name
     */
    static String javaToZapi(String java) {
        return java.replaceAll("(?:([a-z])([A-Z0-9])|([A-Z])([0-9]))", "$1-$2").toLowerCase();
    }
}
